###############################################################################
# (c) Copyright 2021 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
""" SandboxHandler is the implementation of the Sandbox service
    in the DISET framework
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import os
import time
import threading
import tempfile

import six
from DIRAC import gLogger, S_OK, S_ERROR
from DIRAC.Core.Utilities.File import mkDir
from DIRAC.Core.DISET.RequestHandler import RequestHandler
from DIRAC.Core.Security import Locations, Properties, X509Certificate
from DIRAC.WorkloadManagementSystem.DB.SandboxMetadataDB import SandboxMetadataDB
from DIRAC.DataManagementSystem.Client.DataManager import DataManager
from DIRAC.DataManagementSystem.Service.StorageElementHandler import getDiskSpace
from DIRAC.RequestManagementSystem.Client.ReqClient import ReqClient
from DIRAC.RequestManagementSystem.Client.Request import Request
from DIRAC.RequestManagementSystem.Client.Operation import Operation
from DIRAC.RequestManagementSystem.Client.File import File
from DIRAC.Resources.Storage.StorageElement import StorageElement


class SandboxStoreHandler(RequestHandler):

  __purgeCount = -1
  __purgeLock = threading.Lock()
  __purgeWorking = False

  @classmethod
  def initializeHandler(cls, serviceInfoDict):
    """ Initialization of DB object
    """
    cls.sandboxDB = SandboxMetadataDB()
    return S_OK()

  def initialize(self):
    self.__backend = self.getCSOption("Backend", "local")
    self.__localSEName = self.getCSOption("LocalSE", "SandboxSE")
    self.__maxUploadBytes = self.getCSOption("MaxSandboxSizeMiB", 10) * 1048576
    if self.__backend.lower() == "local" or self.__backend == self.__localSEName:
      self.__useLocalStorage = True
      self.__seNameToUse = self.__localSEName
    else:
      self.__useLocalStorage = False
      self.__externalSEName = self.__backend
      self.__seNameToUse = self.__backend
    # Execute the purge once every 1000 calls
    SandboxStoreHandler.__purgeCount += 1
    if SandboxStoreHandler.__purgeCount > self.getCSOption("QueriesBeforePurge", 1000):
      SandboxStoreHandler.__purgeCount = 0
    if SandboxStoreHandler.__purgeCount == 0:
      threading.Thread(target=self.purgeUnusedSandboxes).start()

  def __getSandboxPath(self, md5):
    """ Generate the sandbox path
    """
    # prefix = self.getCSOption( "SandboxPrefix", "SandBox" )
    prefix = "SandBox"
    credDict = self.getRemoteCredentials()
    if Properties.JOB_SHARING in credDict['properties']:
      idField = credDict['group']
    else:
      idField = "%s.%s" % (credDict['username'], credDict['group'])
    pathItems = ["/", prefix, idField[0], idField]
    pathItems.extend([md5[0:3], md5[3:6], md5])
    return os.path.join(*pathItems)

  def transfer_fromClient(self, fileId, token, fileSize, fileHelper):
    """
    Receive a file as a sandbox
    """

    if self.__maxUploadBytes and fileSize > self.__maxUploadBytes:
      fileHelper.markAsTransferred()
      return S_ERROR("Sandbox is too big. Please upload it to a grid storage element")

    if isinstance(fileId, (list, tuple)):
      if len(fileId) > 1:
        assignTo = fileId[1]
        fileId = fileId[0]
      else:
        return S_ERROR("File identified tuple has to have length greater than 1")
    else:
      assignTo = {}

    extPos = fileId.find(".tar")
    if extPos > -1:
      extension = fileId[extPos + 1:]
      aHash = fileId[:extPos]
    else:
      extension = ""
      aHash = fileId
    gLogger.info("Upload requested for %s [%s]" % (aHash, extension))

    credDict = self.getRemoteCredentials()
    sbPath = self.__getSandboxPath("%s.%s" % (aHash, extension))
    # Generate the location
    result = self.__generateLocation(sbPath)
    if not result['OK']:
      return result
    seName, sePFN = result['Value']

    result = self.sandboxDB.getSandboxId(seName, sePFN, credDict['username'], credDict['group'])
    if result['OK']:
      gLogger.info("Sandbox already exists. Skipping upload")
      fileHelper.markAsTransferred()
      sbURL = "SB:%s|%s" % (seName, sePFN)
      assignTo = dict([(key, [(sbURL, assignTo[key])]) for key in assignTo])
      result = self.export_assignSandboxesToEntities(assignTo)
      if not result['OK']:
        return result
      return S_OK(sbURL)

    if self.__useLocalStorage:
      hdPath = self.__sbToHDPath(sbPath)
    else:
      hdPath = False
    # Write to local file
    result = self.__networkToFile(fileHelper, hdPath)
    if not result['OK']:
      gLogger.error("Error while receiving sandbox file", "%s" % result['Message'])
      return result
    hdPath = result['Value']
    gLogger.info("Wrote sandbox to file %s" % hdPath)
    # Check hash!
    if fileHelper.getHash() != aHash:
      self.__secureUnlinkFile(hdPath)
      gLogger.error("Hashes don't match! Client defined hash is different with received data hash!")
      return S_ERROR("Hashes don't match!")
    # If using remote storage, copy there!
    if not self.__useLocalStorage:
      gLogger.info("Uploading sandbox to external storage")
      result = self.__copyToExternalSE(hdPath, sbPath)
      self.__secureUnlinkFile(hdPath)
      if not result['OK']:
        return result
      sbPath = result['Value'][1]
    # Register!
    gLogger.info("Registering sandbox in the DB with", "SB:%s|%s" % (self.__seNameToUse, sbPath))
    result = self.sandboxDB.registerAndGetSandbox(
        credDict['username'], credDict['DN'], credDict['group'],
        self.__seNameToUse, sbPath, fileHelper.getTransferedBytes())
    if not result['OK']:
      self.__secureUnlinkFile(hdPath)
      return result

    sbURL = "SB:%s|%s" % (self.__seNameToUse, sbPath)
    assignTo = dict([(key, [(sbURL, assignTo[key])]) for key in assignTo])
    result = self.export_assignSandboxesToEntities(assignTo)
    if not result['OK']:
      return result
    return S_OK(sbURL)

  def transfer_bulkFromClient(self, fileId, token, _fileSize, fileHelper):
    """ Receive files packed into a tar archive by the fileHelper logic.
        token is used for access rights confirmation.
    """
    result = self.__networkToFile(fileHelper)
    if not result['OK']:
      return result
    tmpFilePath = result['OK']
    gLogger.info("Got Sandbox to local storage", tmpFilePath)

    extension = fileId[fileId.find(".tar") + 1:]
    sbPath = "%s.%s" % (self.__getSandboxPath(fileHelper.getHash()), extension)
    gLogger.info("Sandbox path will be", sbPath)
    # Generate the location
    result = self.__generateLocation(sbPath)
    if not result['OK']:
      return result
    seName, sePFN = result['Value']
    # Register in DB
    credDict = self.getRemoteCredentials()
    result = self.sandboxDB.getSandboxId(
        seName, sePFN, credDict['username'], credDict['group'])
    if result['OK']:
      return S_OK("SB:%s|%s" % (seName, sePFN))

    result = self.sandboxDB.registerAndGetSandbox(
        credDict['username'], credDict['DN'], credDict['group'],
        seName, sePFN, fileHelper.getTransferedBytes())
    if not result['OK']:
      self.__secureUnlinkFile(tmpFilePath)
      return result
    sbid, _newSandbox = result['Value']
    gLogger.info("Registered in DB", "with SBId %s" % sbid)

    result = self.__moveToFinalLocation(tmpFilePath, sbPath)
    self.__secureUnlinkFile(tmpFilePath)
    if not result['OK']:
      gLogger.error("Could not move sandbox to final destination", result['Message'])
      return result

    gLogger.info("Moved to final destination")

    # Unlink temporal file if it's there
    self.__secureUnlinkFile(tmpFilePath)
    return S_OK("SB:%s|%s" % (seName, sePFN))

  def __generateLocation(self, sbPath):
    """
    Generate the location string
    """
    if self.__useLocalStorage:
      return S_OK((self.__localSEName, sbPath))
    # It's external storage
    storageElement = StorageElement(self.__externalSEName)
    res = storageElement.isValid()
    if not res['OK']:
      errStr = "Failed to instantiate destination StorageElement"
      gLogger.error(errStr, self.__externalSEName)
      return S_ERROR(errStr)
    result = storageElement.getURL(sbPath)
    if not result['OK'] or sbPath not in result['Value']['Successful']:
      errStr = "Failed to generate PFN"
      gLogger.error(errStr, self.__externalSEName)
      return S_ERROR(errStr)
    destPfn = result['Value']['Successful'][sbPath]
    return S_OK((self.__externalSEName, destPfn))

  def __sbToHDPath(self, sbPath):
    oldPath = self.__sbToHDPathOld(sbPath)
    newPath = self.__sbToHDPathNew(sbPath)
    if os.path.exists(oldPath):
      return oldPath
    else:
      return newPath

  def __sbToHDPathOld(self, sbPath):
    while sbPath and sbPath[0] == "/":
      sbPath = sbPath[1:]
    oldBasePath = self.getCSOption("BasePath", "/opt/dirac/storage/sandboxes")
    return os.path.join(oldBasePath, sbPath)

  def __sbToHDPathNew(self, sbPath):
    while sbPath and sbPath[0] == "/":
      sbPath = sbPath[1:]
    newBasePath = "/opt/dirac/sandboxStore"
    return os.path.join(newBasePath, sbPath)

  def __networkToFile(self, fileHelper, destFileName=False):
    """
    Dump incoming network data to temporal file
    """
    tfd = None
    if not destFileName:
      try:
        tfd, destFileName = tempfile.mkstemp(prefix="DSB.")
        tfd.close()
      except Exception as e:
        gLogger.error("%s" % repr(e).replace(',)', ')'))
        return S_ERROR("Cannot create temporary file")

    destFileName = os.path.realpath(destFileName)
    mkDir(os.path.dirname(destFileName))

    try:
      if tfd is not None:
        fd = tfd
      else:
        fd = open(destFileName, "wb")
      result = fileHelper.networkToDataSink(fd, maxFileSize=self.__maxUploadBytes)
      fd.close()
    except Exception as e:
      gLogger.error("Cannot open to write destination file", "%s: %s" % (destFileName, repr(e).replace(',)', ')')))
      return S_ERROR("Cannot open to write destination file")
    if not result['OK']:
      return result
    return S_OK(destFileName)

  def __secureUnlinkFile(self, filePath):
    try:
      os.unlink(filePath)
    except Exception as e:
      gLogger.warn("Could not unlink file %s: %s" % (filePath, repr(e).replace(',)', ')')))
      return False
    return True

  def __moveToFinalLocation(self, localFilePath, sbPath):
    if self.__useLocalStorage:
      hdFilePath = self.__sbToHDPath(sbPath)
      result = S_OK((self.__localSEName, sbPath))
      if os.path.isfile(hdFilePath):
        gLogger.info("There was already a sandbox with that name, skipping copy", sbPath)
      else:
        hdDirPath = os.path.dirname(hdFilePath)
        mkDir(hdDirPath)
        try:
          os.rename(localFilePath, hdFilePath)
        except OSError as e:
          errMsg = "Cannot move temporal file to final path"
          gLogger.error(errMsg, repr(e).replace(',)', ')'))
          result = S_ERROR(errMsg)
    else:
      result = self.__copyToExternalSE(localFilePath, sbPath)

    return result

  def __copyToExternalSE(self, localFilePath, sbPath):
    """
    Copy uploaded file to external SE
    """
    try:
      dm = DataManager()
      result = dm.put(sbPath, localFilePath, self.__externalSEName)
      if not result['OK']:
        return result
      if 'Successful' not in result['Value']:
        gLogger.verbose("Oops, no successful transfers there", str(result))
        return S_ERROR("RM returned OK to the action but no successful transfers were there")
      okTrans = result['Value']['Successful']
      if sbPath not in okTrans:
        gLogger.verbose("Ooops, SB transfer wasn't in the successful ones", str(result))
        return S_ERROR("RM returned OK to the action but SB transfer wasn't in the successful ones")
      return S_OK((self.__externalSEName, okTrans[sbPath]))
    except Exception as e:
      gLogger.error("Error while moving sandbox to SE", "%s" % repr(e).replace(',)', ')'))
      return S_ERROR("Error while moving sandbox to SE")

  ##################
  # Assigning sbs to jobs

  types_assignSandboxesToEntities = [dict]

  def export_assignSandboxesToEntities(self, enDict, ownerName="", ownerGroup="", entitySetup=False):
    """
    Assign sandboxes to jobs.
    Expects a dict of { entityId : [ ( SB, SBType ), ... ] }
    """
    if not entitySetup:
      entitySetup = self.serviceInfoDict['clientSetup']
    credDict = self.getRemoteCredentials()
    return self.sandboxDB.assignSandboxesToEntities(
        enDict, credDict['username'], credDict['group'], entitySetup,
        ownerName, ownerGroup)

  ##################
  # Unassign sbs to jobs

  types_unassignEntities = [(list, tuple)]

  def export_unassignEntities(self, entitiesList, entitiesSetup=False):
    """
    Unassign a list of jobs
    """
    if not entitiesSetup:
      entitiesSetup = self.serviceInfoDict['clientSetup']
    credDict = self.getRemoteCredentials()
    return self.sandboxDB.unassignEntities(
        {entitiesSetup: entitiesList}, credDict['username'], credDict['group'])

  ##################
  # Getting assigned sandboxes

  types_getSandboxesAssignedToEntity = [six.string_types]

  def export_getSandboxesAssignedToEntity(self, entityId, entitySetup=False):
    """
    Get the sandboxes associated to a job and the association type
    """
    if not entitySetup:
      entitySetup = self.serviceInfoDict['clientSetup']
    credDict = self.getRemoteCredentials()
    result = self.sandboxDB.getSandboxesAssignedToEntity(
        entityId, entitySetup,
        credDict['username'], credDict['group'])
    if not result['OK']:
      return result
    sbDict = {}
    for SEName, SEPFN, SBType in result['Value']:  # pylint: disable=invalid-name
      if SBType not in sbDict:
        sbDict[SBType] = []
      sbDict[SBType].append("SB:%s|%s" % (SEName, SEPFN))
    return S_OK(sbDict)

  ##################
  # Disk space left management

  types_getFreeDiskSpace = []

  def export_getFreeDiskSpace(self):
    """ Get the free disk space of the storage element
        If no size is specified, terabytes will be used by default.
    """

    return getDiskSpace(
        self.getCSOption("BasePath", "/opt/dirac/storage/sandboxes"))

  types_getTotalDiskSpace = []

  def export_getTotalDiskSpace(self):
    """ Get the total disk space of the storage element
        If no size is specified, terabytes will be used by default.
    """
    return getDiskSpace(
        self.getCSOption("BasePath", "/opt/dirac/storage/sandboxes"), total=True)

  ##################
  # Download sandboxes

  def transfer_toClient(self, fileID, token, fileHelper):
    """ Method to send files to clients.
        fileID is the local file name in the SE.
        token is used for access rights confirmation.
    """
    credDict = self.getRemoteCredentials()
    serviceURL = self.serviceInfoDict['URL']
    filePath = fileID.replace(serviceURL, '')
    result = self.sandboxDB.getSandboxId(self.__localSEName, filePath, credDict['username'], credDict['group'])
    if not result['OK']:
      return result
    sbId = result['Value']
    self.sandboxDB.accessedSandboxById(sbId)
    # If it's a local file
    hdPath = self.__sbToHDPath(filePath)
    if not os.path.isfile(hdPath):
      return S_ERROR("Sandbox does not exist")
    result = fileHelper.getFileDescriptor(hdPath, 'rb')
    if not result['OK']:
      return S_ERROR('Failed to get file descriptor: %s' % result['Message'])
    fd = result['Value']
    result = fileHelper.FDToNetwork(fd)
    fileHelper.oFile.close()
    return result

  ##################
  # Purge sandboxes

  def purgeUnusedSandboxes(self):
    # If a purge is already working skip
    SandboxStoreHandler.__purgeLock.acquire()
    try:
      if SandboxStoreHandler.__purgeWorking:
        if time.time() - SandboxStoreHandler.__purgeWorking < 86400:
          gLogger.info("Sandbox purge still working")
          return S_OK()
      SandboxStoreHandler.__purgeWorking = time.time()
    finally:
      SandboxStoreHandler.__purgeLock.release()

    gLogger.info("Purging sandboxes")
    result = self.sandboxDB.getUnusedSandboxes()
    if not result['OK']:
      gLogger.error("Error while retrieving sandboxes to purge", result['Message'])
      SandboxStoreHandler.__purgeWorking = False
      return result
    sbList = result['Value']
    gLogger.info("Got sandboxes to purge", "(%d)" % len(sbList))
    for sbId, SEName, SEPFN in sbList:  # pylint: disable=invalid-name
      self.__purgeSandbox(sbId, SEName, SEPFN)

    SandboxStoreHandler.__purgeWorking = False
    return S_OK()

  def __purgeSandbox(self, sbId, SEName, SEPFN):
    result = self.__deleteSandboxFromBackend(SEName, SEPFN, useOld=True)
    if not result['OK']:
      gLogger.error("Cannot delete sandbox from backend", result['Message'])
      return
    result = self.__deleteSandboxFromBackend(SEName, SEPFN, useOld=False)
    if not result['OK']:
      gLogger.error("Cannot delete sandbox from backend", result['Message'])
      return
    result = self.sandboxDB.deleteSandboxes([sbId])
    if not result['OK']:
      gLogger.error("Cannot delete sandbox from DB", result['Message'])

  def __deleteSandboxFromBackend(self, SEName, SEPFN, useOld):
    gLogger.info("Purging sandbox" "SB:%s|%s" % (SEName, SEPFN))
    if SEName != self.__localSEName:
      return self.__deleteSandboxFromExternalBackend(SEName, SEPFN)
    else:
      if useOld:
        hdPath = self.__sbToHDPathOld(SEPFN)
      else:
        hdPath = self.__sbToHDPathNew(SEPFN)
      try:
        if not os.path.isfile(hdPath):
          return S_OK()
      except Exception as e:
        gLogger.error("Cannot perform isfile", "%s : %s" % (hdPath, repr(e).replace(',)', ')')))
        return S_ERROR("Error checking %s" % hdPath)
      try:
        os.unlink(hdPath)
      except Exception as e:
        gLogger.error("Cannot delete local sandbox", "%s : %s" % (hdPath, repr(e).replace(',)', ')')))
      while hdPath:
        hdPath = os.path.dirname(hdPath)
        gLogger.info("Checking if dir is empty", hdPath)
        try:
          if not os.path.isdir(hdPath):
            break
          if os.listdir(hdPath):
            break
          gLogger.info("Trying to clean dir", hdPath)
          # Empty dir!
          os.rmdir(hdPath)
        except Exception as e:
          gLogger.error("Cannot clean directory", "%s : %s" % (hdPath, repr(e).replace(',)', ')')))
          break
    return S_OK()

  def __deleteSandboxFromExternalBackend(self, SEName, SEPFN):
    if self.getCSOption("DelayedExternalDeletion", True):
      gLogger.info("Setting deletion request")
      try:

        # We need the hostDN used in order to pass these credentials to the
        # SandboxStoreDB..
        hostCertLocation, _ = Locations.getHostCertificateAndKeyLocation()
        hostCert = X509Certificate.X509Certificate()
        hostCert.loadFromFile(hostCertLocation)
        hostDN = hostCert.getSubjectDN().get('Value')

        # use the host authentication to fetch the data
        result = self.sandboxDB.getSandboxOwner(SEName, SEPFN, hostDN, 'hosts')
        if not result['OK']:
          return result
        _owner, ownerDN, ownerGroup = result['Value']

        request = Request()
        request.RequestName = "RemoteSBDeletion:%s|%s:%s" % (SEName, SEPFN, time.time())
        request.OwnerDN = ownerDN
        request.OwnerGroup = ownerGroup
        physicalRemoval = Operation()
        physicalRemoval.Type = "PhysicalRemoval"
        physicalRemoval.TargetSE = SEName
        fileToRemove = File()
        fileToRemove.PFN = SEPFN
        physicalRemoval.addFile(fileToRemove)
        request.addOperation(physicalRemoval)
        return ReqClient().putRequest(request)
      except Exception as e:
        gLogger.exception("Exception while setting deletion request")
        return S_ERROR("Cannot set deletion request: %s" % str(e))
    else:
      gLogger.info("Deleting external Sandbox")
      try:
        return StorageElement(SEName).removeFile(SEPFN)
      except Exception as e:
        gLogger.exception("RM raised an exception while trying to delete a remote sandbox")
        return S_ERROR("RM raised an exception while trying to delete a remote sandbox")



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""DISET request handler for the LHCbDIRAC/TransformationDB."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import six

from DIRAC import S_OK, S_ERROR
from DIRAC.TransformationSystem.Service.TransformationManagerHandler import TransformationManagerHandler as TManagerBase
from LHCbDIRAC.TransformationSystem.DB.TransformationDB import TransformationDB


class TransformationManagerHandler(TManagerBase):
  types_deleteTransformation = [six.integer_types]

  @classmethod
  def initializeHandler(cls, serviceInfoDict):
    cls.database = TransformationDB('TransformationDB', 'Transformation/TransformationDB')
    return S_OK()

  def export_deleteTransformation(self, transID):
    rc = self.getRemoteCredentials()
    author = rc.get('DN', rc.get('CN'))
    return self.database.deleteTransformation(transID, author=author)

  types_setHotFlag = [six.integer_types, bool]

  def export_setHotFlag(self, transID, hotFlag):
    return self.database.setHotFlag(transID, hotFlag)

  #############################################################################
  #
  # Managing the BkQueries table
  #

  types_addBookkeepingQuery = [six.integer_types, dict]

  @classmethod
  def export_addBookkeepingQuery(self, transID, queryDict):
    return self.database.addBookkeepingQuery(transID, queryDict)

  types_deleteBookkeepingQuery = [six.integer_types]

  @classmethod
  def export_deleteBookkeepingQuery(self, transID):
    return self.database.deleteBookkeepingQuery(transID)

  types_getBookkeepingQuery = [six.integer_types]

  @classmethod
  def export_getBookkeepingQuery(self, transID):
    return self.database.getBookkeepingQuery(transID)

  types_getTransformationsWithBkQueries = [list]

  @classmethod
  def export_getTransformationsWithBkQueries(self, transIDs):
    return self.database.getTransformationsWithBkQueries(transIDs)

  types_setBookkeepingQueryEndRun = [six.integer_types, six.integer_types]

  @classmethod
  def export_setBookkeepingQueryEndRun(self, transID, runNumber):
    return self.database.setBookkeepingQueryEndRun(transID, runNumber)

  types_setBookkeepingQueryStartRun = [six.integer_types, six.integer_types]

  @classmethod
  def export_setBookkeepingQueryStartRun(self, transID, runNumber):
    return self.database.setBookkeepingQueryStartRun(transID, runNumber)

  types_addBookkeepingQueryRunList = [six.integer_types, [list]]

  @classmethod
  def export_addBookkeepingQueryRunList(self, transID, runList):
    return self.database.addBookkeepingQueryRunList(transID, runList)

  #############################################################################
  #
  # Managing the TransformationRuns table
  #

  types_getTransformationRuns = []

  @classmethod
  def export_getTransformationRuns(self, condDict={}, orderAttribute=None, limit=None):
    return self.database.getTransformationRuns(condDict, orderAttribute=orderAttribute, limit=limit)

  types_insertTransformationRun = [six.integer_types, six.integer_types, six.string_types]

  @classmethod
  def export_insertTransformationRun(self, transID, runID, selectedSite=''):
    return self.database.insertTransformationRun(transID, runID, selectedSite='')

  types_getTransformationRunStats = [[six.integer_types, list]]

  @classmethod
  def export_getTransformationRunStats(self, transIDs):
    if isinstance(transIDs, six.integer_types):
      transIDs = [transIDs]
    return self.database.getTransformationRunStats(transIDs)

  types_addTransformationRunFiles = [six.integer_types, six.integer_types, list]

  @classmethod
  def export_addTransformationRunFiles(self, transID, runID, lfns):
    return self.database.addTransformationRunFiles(transID, runID, lfns)

  types_setParameterToTransformationFiles = [six.integer_types, dict]

  @classmethod
  def export_setParameterToTransformationFiles(self, transID, lfnsDict):
    return self.database.setParameterToTransformationFiles(transID, lfnsDict)

  types_setTransformationRunStatus = [six.integer_types, [six.integer_types, list], six.string_types]

  @classmethod
  def export_setTransformationRunStatus(self, transID, runID, status):
    return self.database.setTransformationRunStatus(transID, runID, status)

  types_setTransformationRunsSite = [six.integer_types, six.integer_types, six.string_types]

  @classmethod
  def export_setTransformationRunsSite(self, transID, runID, assignedSE):
    return self.database.setTransformationRunsSite(transID, runID, assignedSE)

  types_getTransformationRunsSummaryWeb = [dict, list, int, int]

  def export_getTransformationRunsSummaryWeb(self, selectDict, sortList, startItem, maxItems):
    """Get the summary of the transformation run information for a given page
    in the generic format."""

    # Obtain the timing information from the selectDict
    last_update = selectDict.get('LastUpdate', None)
    if last_update:
      del selectDict['LastUpdate']
    fromDate = selectDict.get('FromDate', None)
    if fromDate:
      del selectDict['FromDate']
    if not fromDate:
      fromDate = last_update
    toDate = selectDict.get('ToDate', None)
    if toDate:
      del selectDict['ToDate']
    # Sorting instructions. Only one for the moment.
    if sortList:
      orderAttribute = sortList[0][0] + ":" + sortList[0][1]
    else:
      orderAttribute = None

    # Get the transformations that match the selection
    res = self.database.getTransformationRuns(condDict=selectDict, older=toDate,
                                              newer=fromDate, orderAttribute=orderAttribute)
    if not res['OK']:
      self.log.error("TransformationManager.getTransformationRuns()", res['Message'])
      return res

    # Prepare the standard structure now within the resultDict dictionary
    resultDict = {}
    trList = res['Records']
    # Create the total records entry
    nTrans = len(trList)
    resultDict['TotalRecords'] = nTrans
    # Create the ParameterNames entry
    paramNames = res['ParameterNames']
    resultDict['ParameterNames'] = list(paramNames)

    # Add the job states to the ParameterNames entry
    # taskStateNames   = ['Created','Running','Submitted','Failed','Waiting','Done','Stalled']
    # resultDict['ParameterNames'] += ['Jobs_'+x for x in taskStateNames]
    # Add the file states to the ParameterNames entry
    fileStateNames = ['PercentProcessed', 'Processed', 'Unused', 'Assigned',
                      'Total', 'Problematic', 'ApplicationCrash', 'MaxReset']
    resultDict['ParameterNames'] += ['Files_' + x for x in fileStateNames]

    # Get the transformations which are within the selected window
    if nTrans == 0:
      return S_OK(resultDict)
    ini = startItem
    last = ini + maxItems
    if ini >= nTrans:
      return S_ERROR('Item number out of range')
    if last > nTrans:
      last = nTrans
    transList = trList[ini:last]
    if not transList:
      return S_OK(resultDict)

    # Obtain the run statistics for the requested transformations
    transIDs = []
    for transRun in transList:
      transRunDict = dict(zip(paramNames, transRun))
      transID = int(transRunDict['TransformationID'])
      if transID not in transIDs:
        transIDs.append(transID)
    res = self.database.getTransformationRunStats(transIDs)
    if not res['OK']:
      return res
    transRunStatusDict = res['Value']

    statusDict = {}
    # Add specific information for each selected transformation/run
    for transRun in transList:
      transRunDict = dict(zip(paramNames, transRun))
      transID = transRunDict['TransformationID']
      runID = transRunDict['RunNumber']
      if transID not in transRunStatusDict or runID not in transRunStatusDict[transID]:
        for state in fileStateNames:
          transRun.append(0)
        continue
      # Update the status counters
      status = transRunDict['Status']
      statusDict[status] = statusDict.setdefault(status, 0) + 1

      # Populate the run file statistics
      fileDict = transRunStatusDict[transID][runID]
      if fileDict['Total'] == 0:
        fileDict['PercentProcessed'] = 0
      else:
        processed = fileDict.get('Processed', 0)
        fileDict['PercentProcessed'] = "%.1f" % (int(processed * 1000. / fileDict['Total']) / 10.)
      for state in fileStateNames:
        if fileDict and state in fileDict:
          transRun.append(fileDict[state])
        else:
          transRun.append(0)

      # Get the statistics on the number of jobs for the transformation
      # res = database.getTransformationTaskRunStats(transID)
      # taskDict = {}
      # if res['OK'] and res['Value']:
      #  taskDict = res['Value']
      # for state in taskStateNames:
      #  if taskDict and taskDict.has_key(state):
      #    trans.append(taskDict[state])
      #  else:
      #    trans.append(0)

    resultDict['Records'] = transList
    resultDict['Extras'] = statusDict
    return S_OK(resultDict)

  #############################################################################
  #
  # Managing the RunsMetadata table
  #

  types_addRunsMetadata = [six.integer_types, dict]

  @classmethod
  def export_addRunsMetadata(self, runID, metadataDict):
    """insert run metadata."""
    return self.database.setRunsMetadata(runID, metadataDict)

  types_updateRunsMetadata = [six.integer_types, dict]

  @classmethod
  def export_updateRunsMetadata(self, runID, metadataDict):
    """insert run metadata."""
    return self.database.updateRunsMetadata(runID, metadataDict)

  types_getRunsMetadata = [[list, six.integer_types]]

  @classmethod
  def export_getRunsMetadata(self, runID):
    """retrieve run metadata."""
    return self.database.getRunsMetadata(runID)

  types_deleteRunsMetadata = [six.integer_types]

  @classmethod
  def export_deleteRunsMetadata(self, runID):
    """delete run metadata."""
    return self.database.deleteRunsMetadata(runID)

  types_getRunsInCache = [dict]

  @classmethod
  def export_getRunsInCache(self, condDict):
    """gets what's in."""
    return self.database.getRunsInCache(condDict)

  #############################################################################
  #
  # Managing the RunDestination table
  #

  types_getDestinationForRun = [[six.integer_types, six.string_types, list]]

  @classmethod
  def export_getDestinationForRun(self, runIDs):
    """retrieve run destination for a single run or a list of runs."""
    if isinstance(runIDs, six.integer_types):
      runIDs = [runIDs]
    if isinstance(runIDs, str):
      runIDs = [int(runIDs)]
    # expecting a list of long integers
    return self.database.getDestinationForRun(runIDs)

  types_setDestinationForRun = [six.integer_types, six.string_types]

  @classmethod
  def export_setDestinationForRun(self, runID, destination):
    """set run destination."""
    return self.database.setDestinationForRun(runID, destination)

  #############################################################################
  #
  # Managing the StoredJobDescription table
  #

  types_addStoredJobDescription = [six.integer_types, six.string_types]

  @classmethod
  def export_addStoredJobDescription(self, transformationID, jobDescription):
    return self.database.addStoredJobDescription(transformationID, jobDescription)

  types_getStoredJobDescription = [six.integer_types]

  @classmethod
  def export_getStoredJobDescription(self, transformationID):
    return self.database.getStoredJobDescription(transformationID)

  types_removeStoredJobDescription = [six.integer_types]

  @classmethod
  def export_removeStoredJobDescription(self, transformationID):
    return self.database.removeStoredJobDescription(transformationID)

  types_getStoredJobDescriptionIDs = []

  @classmethod
  def export_getStoredJobDescriptionIDs(self):
    return self.database.getStoredJobDescriptionIDs()



#! /usr/bin/env python
###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Set the visibility flag to a dataset."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

from DIRAC import exit as dExit, gLogger
from DIRAC.Core.Utilities.DIRACScript import DIRACScript


@DIRACScript()
def main():
  from LHCbDIRAC.DataManagementSystem.Client.DMScript import DMScript, Script

  dmScript = DMScript()
  dmScript.registerBKSwitches()
  dmScript.registerFileSwitches()

  Script.registerSwitch('', 'List', '   Print out the list of LFNs')
  Script.setUsageMessage(__doc__ + '\n'.join([
      'Usage:',
      '  %s [option|cfgfile]' % Script.scriptName, ]))

  Script.parseCommandLine(ignoreErrors=False)
  dumpList = False
  for switch in Script.getUnprocessedSwitches():
    if switch[0] == 'List':
      dumpList = True

  bkQuery = dmScript.getBKQuery()
  lfns = dmScript.getOption('LFNs', [])
  if not bkQuery and not lfns:
    gLogger.error("No BKQuery and no files given...")
    dExit(1)
  # Invert the visibility flag as want to set Invisible those that are visible and vice-versa
  from LHCbDIRAC.BookkeepingSystem.Client.BookkeepingClient import BookkeepingClient
  bk = BookkeepingClient()

  visibilityFlag = dmScript.getOption('Visibility', None)
  if visibilityFlag is None:
    gLogger.error('Visibility option should be given')
    dExit(2)
  visibilityFlag = str(visibilityFlag).lower() == 'yes'
  if bkQuery:
    # Query with visibility opposite to what is requested to be set ;-)
    bkQuery.setOption('Visible', 'No' if visibilityFlag else 'Yes')
    gLogger.notice("BQ query:", bkQuery)
    lfns += bkQuery.getLFNs()
  if not lfns:
    gLogger.notice("No files found...")
  else:
    res = {'OK': True}
    if visibilityFlag:
      res = bk.setFilesVisible(lfns)
      msg = 'visible'
    else:
      res = bk.setFilesInvisible(lfns)
      msg = 'invisible'
    if not res['OK']:
      gLogger.error("Error setting files", msg)
      dExit(1)
    gLogger.notice("Successfully set %d files %s" % (len(lfns),
                                                     msg + (':' if dumpList else '')))
    if dumpList:
      gLogger.notice('\n'.join(lfns))


if __name__ == "__main__":
  main()



#!/usr/bin/env python
###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Does a TS -> BK check for processed files with descendants."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import sys
import os
import time

import DIRAC
from DIRAC import gLogger
from DIRAC.Core.Utilities.DIRACScript import DIRACScript


@DIRACScript()
def main():
  from DIRAC.Core.Base import Script
  from LHCbDIRAC.DataManagementSystem.Client.DMScript import DMScript

  dmScript = DMScript()
  dmScript.registerFileSwitches()
  depth = 1

  Script.setUsageMessage('\n'.join([__doc__,
                                    'Usage:',
                                    '  %s [option|cfgfile] [ProdIDs]' % Script.scriptName, ]))
  Script.registerSwitch('', 'Runs=', '   Specify the run range')
  Script.registerSwitch('', 'ActiveRunsProduction=', '   Specify the production from which the runs should be derived')
  Script.registerSwitch('', 'FileType=', 'S   pecify the descendants file type')
  Script.registerSwitch('', 'NoFC', '   Trust the BK replica flag, no LFC check')
  Script.registerSwitch('', 'FixIt', '   Fix the files in transformation table')
  Script.registerSwitch('', 'Verbose', '   Print full list of files with error')
  Script.registerSwitch('', 'Status=', '   Select files with a given status in the production')
  Script.registerSwitch('', 'Depth=', '   Depth to which to check descendants (default=%d)' % depth)
  Script.registerSwitch('', 'Force', '   Use this flag to force checking a whole production')
  Script.parseCommandLine(ignoreErrors=True)
  fileType = []
  runsList = []
  lfnList = []
  fixIt = False
  fromProd = None
  verbose = False
  status = None
  noFC = False
  force = False
  for switch in Script.getUnprocessedSwitches():
    if switch[0] == 'Runs':
      try:
        runs = switch[1].split(',')
        runsList = []
        for run in runs:
          runRange = run.split(':')
          if len(runRange) == 2:
            runsList += list(range(int(runRange[0]), int(runRange[1]) + 1))
          else:
            runsList.append(int(run))
      except Exception as e:
        gLogger.exception("Bad run range", switch[1], lException=e)
        DIRAC.exit(1)
    elif switch[0] == 'Status':
      status = switch[1].split(',')
    elif switch[0] == 'Verbose':
      verbose = True
    elif switch[0] == 'FileType':
      fileType = switch[1].split(',')
    elif switch[0] == 'FixIt':
      fixIt = True
    elif switch[0] == 'NoFC':
      noFC = True
    elif switch[0] == 'Depth':
      depth = min(10, max(1, int(switch[1])))
    elif switch[0] == 'Force':
      force = True
    elif switch[0] == 'ActiveRunsProduction':
      try:
        fromProd = int(switch[1])
      except ValueError as e:
        gLogger.exception("Wrong production number: %s" % switch[1], lException=e)
        DIRAC.exit(0)

  args = Script.getPositionalArgs()
  if not len(args):
    gLogger.error("Specify transformation number...")
    DIRAC.exit(0)
  else:
    ids = args[0].split(",")
    prodList = []
    try:
      for id in ids:
        r = id.split(':')
        if len(r) > 1:
          for i in range(int(r[0]), int(r[1]) + 1):
            prodList.append(i)
        else:
          prodList.append(int(eval(r[0])))
    except (ValueError, NameError) as e:
      gLogger.exception("Bad production list: %s" % args[0], lException=e)
  # In case the user asked for specific LFNs
  if not status:
    lfnList = dmScript.getOption('LFNs', [])

  if not status and not lfnList and not runsList and not fromProd and not force:
    gLogger.fatal("You are about to check descendants for all files in a production")
    gLogger.fatal("If you really want to do so, use --Force")
    DIRAC.exit(0)

  from LHCbDIRAC.DataManagementSystem.Client.ConsistencyChecks import ConsistencyChecks
  from LHCbDIRAC.BookkeepingSystem.Client.BKQuery import BKQuery
  from DIRAC.TransformationSystem.Client.TransformationClient import TransformationClient
  tr = TransformationClient()
  for prod in prodList:
    startTime = time.time()
    cc = ConsistencyChecks()
    # Setting the prod also sets its type
    try:
      cc.prod = prod
    except RuntimeError as e:
      gLogger.exception(lException=e)
      continue
    if fileType and cc.transType in ('Merge', 'MCMerge'):
      gLogger.notice("It is not allowed to select file type for merging transformation", prod)
      continue
    cc.verbose = verbose
    cc.noFC = noFC
    cc.descendantsDepth = depth
    if prod != prodList[0]:
      gLogger.notice("====================")
    gLogger.notice("Processing %s production %d" % (cc.transType, cc.prod))

    if status:
      cc.status = status
    if lfnList:
      cc.lfns = lfnList
    if not fileType:
      bkQuery = BKQuery({'Production': prod, 'FileType': 'ALL', 'Visible': 'All'})
      cc.fileType = bkQuery.getBKFileTypes()
      gLogger.notice("Looking for descendants of type %s" % str(cc.fileType))
      notAllFileTypes = False
    else:
      cc.fileType = fileType
      cc.fileTypesExcluded = ['LOG']
      notAllFileTypes = True
    cc.runsList = runsList
    cc.runStatus = 'Active'
    cc.fromProd = fromProd
    cc.checkTS2BK()

    # Print out the results
    gLogger.notice('\nResults:')
    if verbose:
      nMax = sys.maxsize
    else:
      nMax = 20
    suffix = ''
    n = 0
    while True:
      fileName = 'CheckDescendantsResults_%s%s.txt' % (str(cc.prod), suffix)
      if not os.path.exists(fileName):
        break
      n += 1
      suffix = '-%d' % n
    fp = None
    if cc.inFCNotInBK:
      lfns = cc.inFCNotInBK
      gLogger.notice("%d descendants were found in FC but don't have replica flag in BK" % len(lfns))
      if not fp:
        fp = open(fileName, 'w')
      fp.write('\nInFCNotInBK '.join([''] + lfns))
      gLogger.notice('First %d files:' % nMax if not verbose and len(lfns) > nMax else 'All files:',
                     '\n'.join([''] + lfns[0:nMax]))
      gLogger.notice(
          "To fix it:   grep InFCNotInBK %s | dirac-dms-check-fc2bkk\n"
          "\tBeware you can then either fix the BK or the FC/SE" %
          fileName
      )

    if cc.inFailover:
      lfns = cc.inFailover
      gLogger.notice("%d descendants were found in Failover and have no replica flag in BK" % len(lfns))
      if not fp:
        fp = open(fileName, 'w')
      fp.write('\nInFailover '.join([''] + lfns))
      gLogger.notice('First %d files:' % nMax if not verbose and len(lfns) > nMax else 'All files:',
                     '\n'.join([''] + lfns[0:nMax]))
      gLogger.notice(
          "You should check whether they are in a failover request by looking "
          "at their job status and in the RMS..."
      )
      gLogger.notice("To list them:     grep InFailover %s" % fileName)

    if cc.inBKNotInFC:
      lfns = cc.inBKNotInFC
      gLogger.notice("%d descendants were found with replica flag in BK but not in FC" % len(lfns))
      if not fp:
        fp = open(fileName, 'w')
      fp.write('\nInBKNotInFC '.join([''] + lfns))
      gLogger.notice('First %d files:' % nMax if not verbose and len(lfns) > nMax else 'All files:',
                     '\n'.join([''] + lfns[0:nMax]))
      gLogger.notice("To try and fix this:    grep InBKNotInFC %s | dirac-dms-check-bkk2fc" % fileName)

    if cc.removedFiles:
      from DIRAC.Core.Utilities.List import breakListIntoChunks
      gLogger.notice(
          "%d input files are Processed, have no descendants but are no longer in the FC\n"
          "  As they cannot be reset Unused, set them Removed" %
          len(cc.removedFiles))
      if not fp:
        fp = open(fileName, 'w')
      fp.write('\nProcNotinFC '.join([''] + cc.removedFiles))
      gLogger.notice('First %d files:' % nMax if not verbose and len(cc.removedFiles) > nMax else 'All files:',
                     '\n'.join([''] + cc.removedFiles[0:nMax]))
      for lfnChunk in breakListIntoChunks(cc.removedFiles, 1000):
        while True:
          res = cc.transClient.setFileStatusForTransformation(cc.prod, 'Removed', lfnChunk, force=True)
          if not res['OK']:
            gLogger.notice('Error setting files Removed, retry...', res['Message'])
          else:
            break
      gLogger.notice("\tFiles successfully set to status Removed")

    gLogger.notice("%d unique daughters found with real descendants" %
                   (len(set(cc.descForPrcdLFNs).union(cc.descForNonPrcdLFNs))))

    if cc.prcdWithMultDesc:
      lfns = sorted(cc.prcdWithMultDesc)
      gLogger.notice("Processed LFNs with multiple descendants (%d) -> ERROR" % len(lfns))
      gLogger.notice('First %d files:' % nMax if not verbose and len(lfns) > nMax else 'All files:',
                     '\n'.join([''] + lfns[0:nMax]))
      if not fp:
        fp = open(fileName, 'w')
      fp.write('\nProcMultDesc '.join([''] + ['%s: %s' % (lfn, str(multi))
                                              for lfn, multi in cc.prcdWithMultDesc.items()]))
      gLogger.notice("I'm not doing anything for them, neither with the 'FixIt' option")
    else:
      gLogger.notice("No processed LFNs with multiple descendants found -> OK!")

    fixItUsed = False
    if cc.prcdWithoutDesc:
      lfns = set(cc.prcdWithoutDesc)
      res = cc.bkClient.getFileMetadata(list(lfns))
      badLfns = set()
      if res['OK']:
        # Check the DQ flag of these files
        badLfns = set(lfn for lfn, meta in res['Value']['Successful'].items() if meta['DataqualityFlag'] == 'BAD')
      if badLfns:
        gLogger.notice("Processed LFNs without descendants (%d)" % len(lfns))
        gLogger.notice(
            "WARNING: %d of these files have a BAD data quality flag and thus may have been removed" %
            len(badLfns))
        lfns -= badLfns
      if fixIt and lfns:
        fixIt = False
        gLogger.notice("Processed LFNs without descendants (%d) -> ERROR!" % len(lfns))
        gLogger.notice("Resetting them 'Unused'")
        res = cc.transClient.setFileStatusForTransformation(prod, 'Unused', list(lfns), force=True)
        if not res['OK']:
          gLogger.notice("Error resetting files to Unused", res['Message'])
      else:
        fixItUsed = True
        if not fp:
          fp = open(fileName, 'w')
        if badLfns:
          fp.write('\nProcButBAD '.join([''] + sorted(badLfns)))
          gLogger.notice("BAD files processed without descendants can be checked using:")
          gLogger.notice("     grep ProcButBAD %s" % fileName)
        if lfns:
          lfns = sorted(lfns)
          gLogger.notice("Processed LFNs without descendants (%d) -> ERROR!" % len(lfns))
          gLogger.notice('First %d files:' % nMax if not verbose and len(lfns) > nMax else 'All files:',
                         '\n'.join([''] + lfns[0:nMax]))
          fp.write('\nProcNoDesc '.join([''] + lfns))
          if notAllFileTypes:
            gLogger.notice("You may want to check those files again for all file types, using:")
            gLogger.notice("     grep ProcNoDesc %s | dirac-production-check-descendants %d" % (fileName, cc.prod))
          if cc.transType in ('DataStripping', 'MCReconstruction', 'MCStripping'):
            nextProd, nextType = cc._findNextProduction()
            if nextProd:
              gLogger.notice("This is a %s production, thus "
                             "the problem may come from the %s production %d" % (cc.transType, nextType, nextProd))
              gLogger.notice("     You may check it with this command:")
              gLogger.notice("     grep ProcNoDesc %s | dirac-bookkeeping-get-file-descendants --All | grep %d |"
                             "dirac-production-check-descendants %d " % (fileName, cc.prod, nextProd))
          gLogger.notice("If you are sure, use --FixIt for resetting those files Unused in TS")
        else:
          gLogger.notice("No files processed without descendants that are not BAD")
    else:
      gLogger.notice("No processed LFNs without descendants found -> OK!")

    if cc.nonPrcdWithMultDesc:
      lfns = sorted(cc.nonPrcdWithMultDesc)
      gLogger.notice("Non processed LFNs with multiple descendants (%d) -> ERROR" % len(lfns))
      if not fp:
        fp = open(fileName, 'w')
      fp.write('\nNotProcMultDesc '.join([''] + lfns))
      gLogger.notice('First %d files:' % nMax if not verbose and len(lfns) > nMax else 'All files:',
                     '\n'.join([''] + lfns[0:nMax]))
      gLogger.notice("I'm not doing anything for them, neither with the 'FixIt' option")
    else:
      gLogger.notice("No non processed LFNs with multiple descendants found -> OK!")

    # fixing, if requested
    if cc.nonPrcdWithDesc:
      lfns = sorted(cc.nonPrcdWithDesc)
      gLogger.notice("There are %d LFNs not marked Processed but that have descendants -> ERROR" % len(lfns))
      gLogger.notice('First %d files:' % nMax if not verbose and len(lfns) > nMax else 'All files:',
                     '\n'.join([''] + lfns[0:nMax]))
      if fixIt:
        fixIt = False
        gLogger.notice("Marking them as 'Processed'")
        cc.transClient.setFileStatusForTransformation(prod, 'Processed', lfns, force=True)
      else:
        if not fp:
          fp = open(fileName, 'w')
        fp.write('\nNotProcWithDesc '.join([''] + lfns))
        if not fixItUsed:
          gLogger.notice("Use --FixIt for setting files Processed in TS")
        else:
          gLogger.notice("To get the list of files and check them again:")
          gLogger.notice("     grep NotProcWithDesc %s | dirac-production-check-descendants" % fileName)
    else:
      gLogger.notice("No non processed LFNs with descendants found -> OK!")
    if fp:
      fp.close()
      gLogger.notice('Complete list of files is in %s' % fileName)
    gLogger.notice("Processed production %d in %.1f seconds" % (cc.prod, time.time() - startTime))


if __name__ == "__main__":
  main()



#!/usr/bin/env python
###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Insert a new set of simulation conditions in the Bookkeeping."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

from builtins import input

import DIRAC
from DIRAC import gLogger
from DIRAC.Core.Utilities.DIRACScript import DIRACScript


@DIRACScript()
def main():
  from DIRAC.Core.Base import Script
  Script.setUsageMessage(__doc__ + '\n'.join([
      'Usage:',
      '  %s [option|cfgfile] ...' % Script.scriptName]))
  Script.parseCommandLine(ignoreErrors=True)

  from LHCbDIRAC.BookkeepingSystem.Client.BookkeepingClient import BookkeepingClient
  bk = BookkeepingClient()

  exitCode = 0

  desc = input("SimDescription: ")
  beamcond = input("BeamCond: ")
  beamEnergy = input("BeamEnergy: ")
  generator = input("Generator: ")
  magneticField = input("MagneticField: ")
  detectorCond = input("DetectorCond: ")
  luminosity = input("Luminosity: ")
  g4settings = input("G4settings: ")
  gLogger.notice('Do you want to add these new simulation conditions? (yes or no)')
  value = input('Choice:')
  choice = value.lower()
  if choice in ['yes', 'y']:
    in_dict = {
        'SimDescription': desc,
        'BeamCond': beamcond,
        'BeamEnergy': beamEnergy,
        'Generator': generator,
        'MagneticField': magneticField,
        'DetectorCond': detectorCond,
        'Luminosity': luminosity,
        'G4settings': g4settings,
        'Visible': 'Y'}
    res = bk.insertSimConditions(in_dict)
    if res['OK']:
      gLogger.notice('The simulation conditions added successfully!')
    else:
      gLogger.error(res['Message'])
      exitCode = 2
  elif choice in ['no', 'n']:
    gLogger.notice('Aborted!')
  else:
    gLogger.notice('Unexpected choice:', value)
    exitCode = 2

  DIRAC.exit(exitCode)


if __name__ == "__main__":
  main()



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""OutputDataPolicy generates the output data that will be created by a
workflow task.

DIRAC assumes an execute() method will exist during usage.
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from DIRAC import gLogger
from LHCbDIRAC.Interfaces.API.LHCbJob import LHCbJob
from LHCbDIRAC.Core.Utilities.ProductionData import preSubmissionLFNs


class OutputDataPolicy(object):
  """class to generate the output Data."""

  def __init__(self, paramDict):
    """Constructor."""
    self.paramDict = paramDict

  def execute(self):
    """main loop."""
    jobDescription = self.paramDict['Job']
    prodID = self.paramDict['TransformationID']
    jobID = self.paramDict['TaskID']

    job = LHCbJob(jobDescription)
    result = preSubmissionLFNs(job._getParameters(), job.workflow.createCode(),
                               productionID=prodID, jobID=jobID)
    if not result['OK']:
      gLogger.error(result)
    return result



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
""" Analyse XMLSummary module and PoolCatalog in order to monitor the files access
    We send data to the accounting (Site -> SE : fail/success)
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

from collections import defaultdict

from DIRAC import S_OK, S_ERROR, gLogger
from DIRAC.AccountingSystem.Client.Types.DataOperation import DataOperation
from DIRAC.Resources.Catalog.PoolXMLCatalog import PoolXMLCatalog
from LHCbDIRAC.Core.Utilities.XMLSummaries import XMLSummary, XMLSummaryError
from LHCbDIRAC.Workflow.Modules.ModuleBase import ModuleBase


class AnalyseFileAccess(ModuleBase):
  """ Analyzing the access with xroot
  """

  def __init__(self, bkClient=None, dm=None):
    """Module initialization.
    """

    self.log = gLogger.getSubLogger('AnalyseFileAccess')
    super(AnalyseFileAccess, self).__init__(self.log, bkClientIn=bkClient, dm=dm)

    self.version = __RCSID__
    self.XMLSummary = ''
    self.XMLSummary_o = None
    self.poolXMLCatName = ''
    self.poolXMLCatName_o = None
    self.dsc = None

  def _resolveInputVariables(self):
    """ By convention any workflow parameters are resolved here.
    """

    super(AnalyseFileAccess, self)._resolveInputVariables()
    super(AnalyseFileAccess, self)._resolveInputStep()

    self.XMLSummary_o = XMLSummary(self.XMLSummary, log=self.log)
    self.dsc = self.workflow_commons['AccountingReport']
    self.poolXMLCatName_o = PoolXMLCatalog(xmlfile=self.poolXMLCatName)

  def execute(self, production_id=None, prod_job_id=None, wms_job_id=None,
              workflowStatus=None, stepStatus=None,
              wf_commons=None, step_commons=None,
              step_number=None, step_id=None):
    """ Main execution method.

        Here we analyse what is written in the XML summary and the pool XML, and send accounting
    """

    try:
      super(AnalyseFileAccess, self).execute(self.version,
                                             production_id, prod_job_id, wms_job_id,
                                             workflowStatus, stepStatus,
                                             wf_commons, step_commons,
                                             step_number, step_id)

      try:
        self._resolveInputVariables()
      except XMLSummaryError as e:
        if e.message == 'XML Summary Not Available':
          self.log.warn("XML summary not created, skipping this module")
          return S_OK()
        else:
          raise e

      self.log.info("Analyzing root access from %s and %s" % (self.XMLSummary, self.poolXMLCatName))

      accessAttempts = self._checkFileAccess(self.poolXMLCatName_o, self.XMLSummary_o)

      if not self._enableModule():
        self.log.info('Not enabled')
        return S_OK()

      for remoteSE, success in accessAttempts:
        oDataOperation = self.__initialiseAccountingObject(remoteSE, success)
        self.dsc.addRegister(oDataOperation)

      return S_OK()

    except Exception as e:  # pylint:disable=broad-except
      self.log.exception("Failure in AnalyseFileAccess execute module", lException=e)
      self.setApplicationStatus(e)
      return S_ERROR(str(e))

    finally:
      super(AnalyseFileAccess, self).finalize(self.version)

  @staticmethod
  def _checkFileAccess(xmlCatalog, xmlSummary):
    """ Given an xmlCatalog and an xmlSummary, check which were the successful and failed attempts
        to open remote root files

        For each attempts, we return a tuple (srcSE, flag) with the flag being true if the read was successful.

        :param xmlCatalog: instance of :py:class:`~LHCbDIRAC.Resources.Catalog.PoolXMLCatalog.PoolXMLCatalog`
        :param xmlSummary: instance of :py:class:`~LHCbDIRAC.LHCbDIRAC.Core.Utilities.XMLSummaries.XMLSummary`

        :returns: list of tuples (srcSE, successful flag)
    """

    # This will contain the list of tuples with the accesses and their status
    accessAttempts = []

    # Used to cache the information PFN -> SE from the catalog
    pfn_se = {}
    # Used to cache the information LFN -> PFN from the catalog
    lfn_pfn = defaultdict(list)

    # Matching between a PFN and the LFN
    pfn_lfn = {}

    # {LFN: failed PFN}
    lfn_pfn_fail = defaultdict(list)

    # Build all the caches
    for guid in xmlCatalog.files:
      pFile = xmlCatalog.files[guid]
      lfn = pFile.lfns[0]  # there can be only one
      for pfn, _ftype, se in pFile.pfns:
        pfn_lfn[pfn] = lfn
        pfn_se[pfn] = se
        lfn_pfn[lfn].append(pfn)

    # Check the URLs that failed, and associate them with their LFNs
    for pfn, _status in xmlSummary.failedInputURL:
      # The name starts with 'PFN:'
      pfn = pfn[4:]
      # Find the matching LFN
      lfn = pfn_lfn[pfn]
      lfn_pfn_fail[lfn].append(pfn)

    # For each PFN in the lfn_pfn_fail dictionnary, we will count a failure.
    for lfn, failedPfns in lfn_pfn_fail.items():
      # We add the accounting for the failure
      for pfn in failedPfns:
        remoteSE = pfn_se[pfn]
        accessAttempts.append((remoteSE, False))

    # To find the successful access, we take the next available replicas if any.
    # If the LFN is not in the lfn_pfn_fail dict, that means the first attempt was successful
    # If there are no more replicas to be tried, then the LFN was never read.

    for lfn, replicaList in lfn_pfn.items():
      # get the index of the successful replicas
      succRepIndex = len(lfn_pfn_fail.get(lfn, []))
      try:
        succRep = replicaList[succRepIndex]
        remoteSE = pfn_se[succRep]
        accessAttempts.append((remoteSE, True))
      # This happens if all the replicas failed
      except IndexError:
        pass

    return accessAttempts

  def __initialiseAccountingObject(self, srcSE, successful):
    """ create accouting record """
    accountingDict = {}

    accountingDict['OperationType'] = 'fileAccess'
    accountingDict['User'] = self._getCurrentOwner()
    accountingDict['Protocol'] = 'xroot'
    accountingDict['RegistrationTime'] = 0.0
    accountingDict['RegistrationOK'] = 0
    accountingDict['RegistrationTotal'] = 0
    accountingDict['Source'] = srcSE
    accountingDict['TransferTotal'] = 1
    accountingDict['TransferOK'] = 1 if successful else 0
    accountingDict['TransferSize'] = 0
    accountingDict['TransferTime'] = 0.0
    accountingDict['FinalStatus'] = 'Successful' if successful else 'Failed'
    accountingDict['Destination'] = self.siteName
    oDataOperation = DataOperation()
    oDataOperation.setValuesFromDict(accountingDict)

    if 'StartTime' in self.step_commons:
      oDataOperation.setStartTime(self.step_commons['StartTime'])
      oDataOperation.setEndTime()

    return oDataOperation



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Reads .log-files and outputs summary of counters as a .json-file and a
.html-file."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
import os
import json

from distutils.version import LooseVersion  # pylint:disable=import-error,no-name-in-module

from DIRAC import gLogger, S_OK, S_ERROR


def readLogFile(logFile, project, version, appConfigVersion, jobID, prodID, wmsID, name='errors.json'):
  """The script that runs everything.

  :param str logFile: the name of the logfile
  :param str project: the project string of the file
  :param str version: the versio of the project
  :param str jobID: the JobID
  :param str prodID: the production ID
  :param str wmsID: the wmsID
  :param str name: the name of the output json file, standardised to 'errors.json'
  """
  logString = ''
  fullPathFileName = pickStringFile(project, version, appConfigVersion)
  dictTotal = []
  dictG4Errors = dict()
  dictG4ErrorsCount = dict()

  if fullPathFileName is None or not os.path.exists(fullPathFileName) or os.stat(fullPathFileName)[6] == 0:
    gLogger.warn('STRINGFILE %s is empty' % fullPathFileName)

  readErrorDict(fullPathFileName, dictG4Errors)

  res = getLogString(logFile, logString)
  if not res['OK']:
    gLogger.warn('Problems in reading %s' % logFile)
    return res
  logString = res['Value']

  reversedKeys = sorted(dictG4Errors, reverse=True)

  for errorString in reversedKeys:
    dictCountDumpErrorString = dict()
    dictTemp = dict()
    ctest = logString.count(errorString)
    test = logString.find(errorString)
    array = []
    for i in range(ctest):
      start = test
      test = logString.find(errorString, start)
      alreadyFound = False
      for error in reversedKeys:
        if error == errorString:
          break
        checke = logString[test:test + 100].find(error)
        if checke != -1:
          alreadyFound = True
          test = test + len(error)
          break
      if alreadyFound:
        continue

      if test != -1:
        eventnr = ''
        runnr = ''

        eventnr_point = logString.rfind('INFO Evt', test - 5000, test)
        if eventnr_point != -1:
          eventnr = 'Evt ' + logString[eventnr_point:test].split('INFO Evt')[1].strip().split(',')[0]
          runnr = logString[eventnr_point:].split('INFO Evt')[1].strip().split(',')[1]

        if errorString.find('G4') != -1:
          check = logString[test:test + 250].find('***')
          if check != -1:
            errorBase = logString[test:test + 250].split('***')[0]
            dictCountDumpErrorString[i] = eventnr + "  " + runnr + "  -->" + errorBase

            array.append(dict())
            array[-1]['eventnr'] = eventnr
            array[-1]['runnr'] = runnr

            lengthDump = len(errorBase)
            test = test + lengthDump
        else:
          errorBase = logString[test:test + 250].split('\n')[0]
          dictCountDumpErrorString[i] = eventnr + "  " + runnr + "  -->" + errorBase

          array.append(dict())
          array[-1]['eventnr'] = eventnr
          array[-1]['runnr'] = runnr

          lengthDump = len(errorBase)
          test = test + lengthDump

        if array[-1] is not None:
          dictTemp[errorString] = dict()
          dictTemp[errorString] = array

    if dictTemp != {}:
      dictTotal.append(dictTemp)
    dictG4ErrorsCount[errorString] = dictCountDumpErrorString

  createJSONtable(dictTotal, name, jobID, prodID, wmsID)
  createHTMLtable(dictG4ErrorsCount, "errors.html")

  return S_OK()

################################################

#   # Due to issues in the mapping of the ES DB, this mapping
#   (which is more clear than the one below) couldn't be used.
#   # I have still saved the function here.

# def createJSONtable(dictTotal, name):

#   ids = {}
#   ids['JobID'] = JOB_ID
#   ids['ProductionID'] = PROD_ID
#   ids['TransformationID'] = TRANS_ID

#   errors = []

#   with open(name, 'w') as output:
#     for error in dictTotal:
#       newrow = {}
#       for key, value in error.items():
#         newrow['Error type'] = key
#         newrow['Counter'] = len(value)
#         newrow['Events'] = value
#       errors.append(newrow)

#     errorDict = {'Errors' : errors}

#     result = {}
#     result['ID'] = ids
#     result['Errors'] = errors
#     result = {'Log_output' : result}

#     output.write(json.dumps(result, indent = 2))
#   return

# \


def createJSONtable(dictTotal, name, jobID, prodID, wmsID):
  """Creates a JSON file out of the collection of errors listed in dictTotal.

  :param dict dictTotal: the dictionary of errors
  :param str name: the name of the JSON file
  :param str jobID: the JobID of the log
  :param str prodID: the ProductionID of the log
  :param str wmsID: the wmsID of the log
  """
  result = {}
  result['JobID'] = jobID
  result['ProductionID'] = prodID
  result['wmsID'] = wmsID

  with open(name, 'w') as output:
    for error in dictTotal:
      for key, value in error.items():
        result[key] = len(value)
    json.dump(result, output, indent=2)

#####################################################


def createHTMLtable(dictG4ErrorsCount, name):
  """Creates an HTML file out of the collection of errors listed in
  dictG4ErrorsCount.

  :param dict dictG4ErrorsCount: the dictionary of errors
  :param str name: the name of the HTML file
  """
  with open(name, 'w') as f:
    f.write("<HTML>\n")

    f.write("<table border=1 bordercolor=#000000 width=100% bgcolor=#BCCDFE>")
    f.write("<tr>")
    f.write("<td>ERROR TYPE</td>")
    f.write("<td>COUNTER</td>")
    f.write("<td>DUMP OF ERROR MESSAGES</td>")
    f.write("</tr>")

    orderedKeys = sorted(dictG4ErrorsCount)
    for errString in orderedKeys:
      if dictG4ErrorsCount[errString] != {}:
        f.write("<tr>")
        f.write("<td>" + errString + "</td>")
        f.write("<td>" + str(len(dictG4ErrorsCount[errString])) + "</td>")
        f.write("<td>")
        f.write("<lu>")
        for y in dictG4ErrorsCount[errString]:
          f.write("<li>")
          f.write(" " + dictG4ErrorsCount[errString][y] + " ")
        f.write("</lu>")
        f.write("</td>")
        f.write("</tr>")

    f.write("</table>")

#######################################################


def readErrorDict(fullPathFileName, dictName):
  """Reads errors in a stringfile and puts them in dictName.

  :param str fullPathFileName: the name of the stringfile
  :param dict dictName: the name of the dict that will insert the data in fullPathFileName
  """
  fileLines = getLines(fullPathFileName)
  for line in fileLines:
    errorString = line.split(',')[0]
    description = line.split(',')[1]
    dictName[errorString] = description

################################################


def getLines(fullPathFileName):
  """Reads lines in string file.

  :param str fullPathFileName: the name of the file to be opened and read
  """

  gLogger.notice('>>> Processed STRINGFILE -> ', fullPathFileName)
  with open(fullPathFileName, 'r') as f:
    lines = f.readlines()
  return lines

################################################


def getLogString(logFile, logString):
  """Checks if the log file can be opened, and saves the text in logFile into
  logString.

  :param str logFile: the name of the logFile
  :param str logStr: the name of the variable that will save the contents of logFile
  """

  gLogger.notice('Attempting to open %s' % logFile)
  if not os.path.exists(logFile):
    gLogger.error('%s could not be found' % logFile)
    return S_ERROR()
  if os.stat(logFile)[6] == 0:
    gLogger.error('%s is empty' % logFile)
    return S_ERROR()
  with open(logFile, 'r') as f:
    logString = f.read()
  gLogger.notice("Successfully read %s" % logFile)
  return S_OK(logString)

################################################


def pickStringFile(project, version, appConfigVersion):
  """Picks the string file from the current directory.

  :param str project: the project name
  :param str version: the version of the project
  :param str version: APPCONFIG version
  """

  if os.environ.get('VO_LHCB_SW_DIR'):
    sharedArea = os.environ['VO_LHCB_SW_DIR']

    # sometimes this is wrong... so trying to correct it!
    if 'lhcb' not in os.listdir(sharedArea):
      sharedArea = os.path.join(sharedArea, 'lib')
      if 'lhcb' not in os.listdir(sharedArea):
        gLogger.error("Current sharedArea (%s) content: " % sharedArea, os.listdir(sharedArea))
        raise RuntimeError("Can't find a sharedArea")

  else:
    sharedArea = '/cvmfs/lhcb.cern.ch/lib'

  sourceDir = os.path.join(sharedArea, 'lhcb', 'DBASE', 'AppConfig', appConfigVersion, 'errstrings')
  fileName = project + '_' + version + '_errs.txt'
  fullPathFileName = os.path.join(sourceDir, os.path.basename(fileName))
  if not os.path.exists(fullPathFileName):
    gLogger.notice('string file %s does not exist, attempting to take the most recent file ...' % fullPathFileName)
    fileList = [fn for fn in os.listdir(sourceDir) if project in fn]
    if fileList:
      versionsList = [x.replace(project + '_', '').replace('_errs.txt', '') for x in fileList]
      mostRecentVersion = sorted(versionsList, key=LooseVersion)[-1]
      fileName = project + '_' + mostRecentVersion + '_errs.txt'
      fullPathFileName = os.path.join(sourceDir, os.path.basename(fileName))
    else:
      gLogger.warn('WARNING: no string files for this project')
      return None

  return fullPathFileName

# This is a relic from the previous version of the file, I'll keep it for the while

# global LOG_STRING
# global STRING_FILE
# global FILE_OK

# LOG_FILE = sys.argv[1]
# PROJECT = sys.argv[2]
# VERSION = sys.argv[3]

# global JOB_ID
# global PROD_ID
# global TRANS_ID

# JOB_ID = sys.argv[4]
# PROD_ID = sys.argv[5]
# TRANS_ID = sys.argv[6]

# #LOG_STRING = ''
# #FILE_OK = ''
# dictG4Errors = dict()
# dictG4ErrorsCount = dict()
# STRING_FILE = pickStringFile(PROJECT, VERSION)

# if STRING_FILE is not None:
#   if os.stat(STRING_FILE)[6] != 0:
#     main(LOG_FILE)
#   else:
#     print 'WARNING: STRINGFILE %s is empty' % STRING_FILE

# The file is run as follows:
# readLogFile('Example.log', 'project', 'version', 'jobID', 'prodID', 'wmsID')



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
""" Module containing a front-end to the ElasticSearch-based ElasticGeneratorLogDB.
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function


__RCSID__ = "$Id$"

from DIRAC import gConfig
from DIRAC.ConfigurationSystem.Client.PathFinder import getDatabaseSection
from DIRAC.ConfigurationSystem.Client.Helpers import CSGlobals
from LHCbDIRAC.ProductionManagementSystem.DB.ElasticMCStatsDBBase import ElasticMCStatsDBBase

name = 'ElasticGeneratorLogDB'

mapping = {
    "properties": {
        "wmsID": {"type": "long"},
        "ProductionID": {"type": "integer"},
        "JobID": {"type": "integer"},
        # TODO: ADD A Mapping
    }
}


class ElasticGeneratorLogDB(ElasticMCStatsDBBase):

  def __init__(self):
    """ Standard Constructor
    """

    section = getDatabaseSection("ProductionManagement/ElasticGeneratorLogDB")
    indexPrefix = gConfig.getValue("%s/IndexPrefix" % section,
                                   CSGlobals.getSetup()).lower()

    # Connecting to the ES cluster
    super(ElasticGeneratorLogDB, self).__init__(name,
                                                'ProductionManagement/ElasticGeneratorLogDB',
                                                indexPrefix)

    self.indexName = "%s_%s" % (self.getIndexPrefix(), name.lower())
    # Verifying if the index is there, and if not create it
    if not self.client.indices.exists(self.indexName):
      result = self.createIndex(self.indexName, mapping, period=None)
      if not result['OK']:
        self.log.error(result['Message'])
        raise RuntimeError(result['Message'])
      self.log.always("Index created:", self.indexName)

    self.dslSearch = self._Search(self.indexName)



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""It is used to test the Bookkeeping utilities."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
import unittest

from LHCbDIRAC.BookkeepingSystem.DB.Utilities import enoughParams, _ONE, _TWO

__RCSID__ = "$Id$"


class UtilitiesTestCase(unittest.TestCase):

  def test_enoughParams(self):

    result = enoughParams({})
    self.assertEqual(result, False)

    result = enoughParams({'ReplicaFlag': 'Test', 'Visible': 'Test'})
    self.assertEqual(result, False)

    for i in _ONE:
      result = enoughParams({i: 'Test'})
      self.assertEqual(result, False)

    for i in _TWO:
      for j in _TWO:
        result = enoughParams({i: 'Test', j: 'Test'})
        self.assertEqual(result, False)

    result = enoughParams({'ConfigName': 'Test', 'ConfigVersion': 'Test', 'Production': 'Test'})
    self.assertEqual(result, True)

    result = enoughParams({'Production': 'Test', 'ReplicaFlag': 'Test', 'Visible': 'Test'})
    self.assertEqual(result, True)


if __name__ == '__main__':

  suite = unittest.defaultTestLoader.loadTestsFromTestCase(UtilitiesTestCase)
  testResult = unittest.TextTestRunner(verbosity=2).run(suite)



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""
:mod: RAWIntegrityAgent

.. module: RAWIntegrityAgent

:synopsis: RAWIntegrityAgent determines whether RAW files in CASTOR were migrated correctly.
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
# # imports
import datetime
# # from DIRAC
from DIRAC import S_OK
# # from Core
from DIRAC.Core.Base.AgentModule import AgentModule
from DIRAC.ConfigurationSystem.Client.ConfigurationData import gConfigurationData
from DIRAC.FrameworkSystem.Client.MonitoringClient import gMonitor
# # from DMS
from DIRAC.Resources.Catalog.FileCatalog import FileCatalog
from DIRAC.Resources.Storage.StorageElement import StorageElement
from LHCbDIRAC.DataManagementSystem.DB.RAWIntegrityDB import RAWIntegrityDB
from DIRAC.Core.Utilities.List import breakListIntoChunks

__RCSID__ = "$Id$"

AGENT_NAME = 'DataManagement/RAWIntegrityAgent'


class RAWIntegrityAgent(AgentModule):
  """.. class:: RAWIntegirtyAgent.

  :param RAWIntegrityDB rawIntegrityDB: RAWIntegrityDB instance
  :param str gatewayUrl: URL to online RequestClient
  """

  def __init__(self, *args, **kwargs):
    """c'tor."""

    AgentModule.__init__(self, *args, **kwargs)

    self.rawIntegrityDB = None
    self.fileCatalog = None
    self.onlineRequestMgr = None

  def initialize(self):
    """agent initialisation."""

    self.rawIntegrityDB = RAWIntegrityDB()

    # The file catalog is used to register file once it has been transfered
    # But we want to register it in all the catalogs except the RAWIntegrityDB
    # otherwise it is register twice
    # We also remove the BK catalog because some files are not registered there
    # (detector calibration files for example). The real data are registered in
    # the bookeeping by the DataMover
    self.fileCatalog = FileCatalog(catalogs='FileCatalog')

    gMonitor.registerActivity("Iteration", "Agent Loops/min", "RAWIntegriryAgent", "Loops",
                              gMonitor.OP_SUM)
    gMonitor.registerActivity("WaitingFiles", "Files waiting for migration", "RAWIntegriryAgent",
                              "Files", gMonitor.OP_MEAN)
    gMonitor.registerActivity("WaitSize", "Size of migration buffer", "RAWIntegrityAgent", "GB",
                              gMonitor.OP_MEAN)

    gMonitor.registerActivity("NewlyMigrated", "Newly migrated files", "RAWIntegriryAgent", "Files",
                              gMonitor.OP_SUM)
    gMonitor.registerActivity("TotMigrated", "Total migrated files", "RAWIntegriryAgent", "Files",
                              gMonitor.OP_ACUM)
    gMonitor.registerActivity("TotMigratedSize", "Total migrated file size", "RAWIntegriryAgent",
                              "GB", gMonitor.OP_ACUM)

    gMonitor.registerActivity("BadChecksum", "Checksums mismatch", "RAWIntegriryAgent", "Files",
                              gMonitor.OP_SUM)
    gMonitor.registerActivity("ErrorMetadata", "Error when getting files metadata",
                              "RAWIntegriryAgent", "Files", gMonitor.OP_SUM)
    gMonitor.registerActivity("ErrorRegister", "Error when registering files", "RAWIntegriryAgent",
                              "Files", gMonitor.OP_SUM)
    gMonitor.registerActivity("ErrorRemove", "Error when removing files", "RAWIntegriryAgent",
                              "Files", gMonitor.OP_SUM)

    gMonitor.registerActivity("FailedMigrated",
                              "Number of files encountering error during migration",
                              "RAWIntegriryAgent", "Files", gMonitor.OP_SUM)
    gMonitor.registerActivity("TotFailMigrated",
                              "Total number of files encountering error during migration",
                              "RAWIntegriryAgent", "Files", gMonitor.OP_ACUM)

    gMonitor.registerActivity("MigrationTime", "Average migration time", "RAWIntegriryAgent",
                              "Seconds", gMonitor.OP_MEAN)
    # gMonitor.registerActivity("TimeInQueue", "Average current wait for migration",
    #                           "RAWIntegriryAgent", "Minutes", gMonitor.OP_MEAN)
    gMonitor.registerActivity("MigrationRate", "Observed migration rate", "RAWIntegriryAgent",
                              "MB/s", gMonitor.OP_MEAN)

    # This sets the Default Proxy to used as that defined under
    # /Operations/Shifter/DataManager
    # the shifterProxy option in the Configuration can be used to change this default.
    self.am_setOption('shifterProxy', 'DataProcessing')

    return S_OK()

  def _checkMigrationStatus(self, filesMetadata, lfnsMetadata):
    """Check that the lfns in parameters are properly migrated, and compares
    the checksum between castor and the RAWIntegrityDB.

    :param filesMetadata: dict {lfn: se metadata}
    :param lfnsMetadata: dict {lfn: metadata in RAWIntegrityDB}

    :returns: True/False in successful, Failed in case of problem
    """
    ############################################################
    #
    # Determine the files that have been newly migrated and their success
    #

    successful = {}
    failed = {}
    for lfn, seMetadata in filesMetadata.items():
      isMigrated = seMetadata.get('Migrated', False)
      # If it is not migrated, go to the next one
      if not isMigrated:
        successful[lfn] = False
        continue
      else:
        self.log.info("%s is copied." % lfn)
        castorChecksum = seMetadata['Checksum']
        onlineChecksum = lfnsMetadata[lfn]['Checksum']
        if castorChecksum.lower().lstrip('0') == onlineChecksum.lower().lstrip('0').lstrip('x'):
          self.log.info("%s migrated checksum match." % lfn)
          successful[lfn] = True
        else:
          errStr = "Migrated checksum mis-match.", "%s %s %s" % (
              lfn, castorChecksum.lstrip('0'), onlineChecksum.lstrip('0').lstrip('x'))
          self.log.error(errStr)
          failed[lfn] = errStr

    return S_OK({'Successful': successful, 'Failed': failed})

  def getNewlyCopiedFiles(self, activeFiles):
    """Get the list of files newly copied and those not yet copied.

    :param activeFiles: dict {lfn:RawIntegrityDB metadata} for files in Active status

    :return: tuple filesNewlyCopied, filesNotYetCopied (lfns lists)
    """
    # This is a list of all the lfns that we will have newly copied
    filesNewlyCopied = []
    # This are the lfns that are not yet copied
    filesNotYetCopied = []

    self.log.info("Obtaining physical file metadata.")
    # Group the lfns by SEs
    seLfns = {}
    for lfn, metadataDict in activeFiles.items():
      se = metadataDict['SE']
      seLfns.setdefault(se, []).append(lfn)

    for se in sorted(seLfns):
      lfnList = seLfns[se]
      failedMetadata = {}
      successfulMetadata = {}
      res = StorageElement(se).getFileMetadata(lfnList)
      if not res['OK']:
        errStr = "Failed to obtain physical file metadata."
        self.log.error(errStr, res['Message'])
        failedMetadata = dict((lfn, errStr) for lfn in lfnList)
      else:
        successfulMetadata = res['Value']['Successful']
        failedMetadata = res['Value']['Failed']

      if failedMetadata:
        self.log.info("Failed to obtain physical file metadata for %s files." % len(failedMetadata))
        gMonitor.addMark('ErrorMetadata', len(failedMetadata))

      if successfulMetadata:
        self.log.info("Obtained physical file metadata for %s files." % len(successfulMetadata))

        ############################################################
        #
        # Determine the files that have been newly migrated and their success
        #
        res = self._checkMigrationStatus(successfulMetadata, activeFiles)
        if not res['OK']:
          self.log.error("Error when checking migration status", res)
          gMonitor.addMark("BadChecksum", len(successfulMetadata))
        else:
          succCompare = res['Value']['Successful']
          failedCompare = res['Value']['Failed']
          seFilesCopied = []
          seFilesNotCopied = []
          # The copied files are those in True in the successful dictionary
          for lfn, isCopied in succCompare.items():
            if isCopied:
              seFilesCopied.append(lfn)
            else:
              seFilesNotCopied.append(lfn)

          filesNewlyCopied.extend(seFilesCopied)
          filesNotYetCopied.extend(seFilesNotCopied)

          gMonitor.addMark("BadChecksum", len(failedCompare))

          self.log.info("%s files newly copied at %s." % (len(seFilesCopied), se))
          self.log.info("Found %s checksum mis-matches at %s." % (len(failedCompare), se))

    return filesNewlyCopied, filesNotYetCopied

  def registerCopiedFiles(self, filesNewlyCopied, copiedFiles, allUnmigratedFilesMeta):
    """Register successfuly copied files (newly, or in Copied status in the DB)
    in the DFC.

    :param filesNewlyCopied: [lfns] of files newly copied
    :param copiedFiles: {lfn:RIDb metadata} of files that were in Copied state.
    :param allUnmigratedFilesMeta: {lfn:RI Db metadata} for all lfns non migrated at
                                  the beginning of the loop.

    :return: {lfn:True} for successfuly registered lfns
    """
    if filesNewlyCopied or copiedFiles:
      self.log.info("Attempting to register %s newly copied and %s previously copied files" %
                    (len(filesNewlyCopied), len(copiedFiles)))
    else:
      self.log.info("No files to be registered")

    # Update copiedFiles to also contain the newly copied files
    copiedFiles.update(dict((lfn, allUnmigratedFilesMeta[lfn]) for lfn in filesNewlyCopied))

    successfulRegister = {}
    failedRegister = {}

    # Try to register them by batch
    for lfnChunk in breakListIntoChunks(copiedFiles, 100):
      # Add the metadata
      lfnDictChuck = dict((lfn, copiedFiles[lfn]) for lfn in lfnChunk)
      res = self.fileCatalog.addFile(lfnDictChuck)

      if not res['OK']:
        self.log.error("Completely failed to register some successfully copied file.",
                       res['Message'])
        failedRegister.update(dict((lfn, res['Message']) for lfn in lfnDictChuck))
      else:
        successfulRegister.update(res['Value']['Successful'])
        failedRegister.update(res['Value']['Failed'])

    gMonitor.addMark("ErrorRegister", len(failedRegister))
    for lfn, reason in failedRegister.items():
      self.log.error("Failed to register lfn. Setting to Copied", "%s: %s" % (lfn, reason))
      res = self.rawIntegrityDB.setFileStatus(lfn, 'Copied')
      if not res['OK']:
        self.log.error("Error setting file status to Copied", "%s: %s" % (lfn, res['Message']))

    for lfn in successfulRegister:
      self.log.info("Successfully registered %s in the File Catalog." % lfn)

    return successfulRegister

  def removeRegisteredFiles(self, filesNewlyRegistered, registeredFiles, allUnmigratedFilesMeta):
    """Remove successfuly registered files (newly, or in Registered status in
    the DB) from the OnlineStorage.

    :param filesNewlyCopied: [lfns] of files newly copied
    :param copiedFiles: {lfn:RIDb metadata} of files that were in Copied state.
    :param allUnmigratedFilesMeta: {lfn:RI Db metadata} for all lfns non migrated at
                                  the beginning of the loop.

    :return: {lfn:True} for successfuly registered lfns
    """
    if filesNewlyRegistered or registeredFiles:
      self.log.info("Attempting to remove %s newly registered and %s previously registered files" %
                    (len(filesNewlyRegistered), len(registeredFiles)))
    else:
      self.log.info("No files to be removed")

    # Update registeredFiles to also contain the newly registered files
    registeredFiles.update(dict((lfn, allUnmigratedFilesMeta[lfn]) for lfn in filesNewlyRegistered))

    onlineSE = StorageElement('OnlineRunDB')

    # Try to them them all
    res = onlineSE.removeFile(registeredFiles)

    filesNewlyRemoved = {}
    failedRemove = {}
    if not res['OK']:
      self.log.error("Completely failed to remove successfully registered files.", res['Message'])
      failedRemove = dict((lfn, res['Message']) for lfn in registeredFiles)
    else:
      filesNewlyRemoved = res['Value']['Successful']
      failedRemove = res['Value']['Failed']

    gMonitor.addMark("ErrorRemove", len(failedRemove))
    for lfn, reason in failedRemove.items():
      self.log.error("Failed to remove lfn. Setting to Registered", "%s: %s" % (lfn, reason))
      res = self.rawIntegrityDB.setFileStatus(lfn, 'Registered')
      if not res['OK']:
        self.log.error("Error setting file status to Registered", "%s: %s" % (lfn, res['Message']))

    now = datetime.datetime.utcnow()
    for lfn in filesNewlyRemoved:
      self.log.info("Successfully removed %s from the Online storage. Setting it to Done" % lfn)
      res = self.rawIntegrityDB.setFileStatus(lfn, 'Done')
      if not res['OK']:
        self.log.error("Error setting file status to Done", "%s: %s" % (lfn, res['Message']))
      else:
        # SubmitTime is ALREADY a datetime since it is declared as such in the DB.
        submitTime = allUnmigratedFilesMeta[lfn]['SubmitTime']
        migrationTime = (now - submitTime).total_seconds()
        gMonitor.addMark("MigrationTime", migrationTime)
        fileSizeMB = allUnmigratedFilesMeta[lfn]['Size'] / (1024 * 1024.0)
        gMonitor.addMark("MigrationRate", fileSizeMB / migrationTime)

    return filesNewlyRemoved

  def execute(self):
    """execution in one cycle."""

    # Don't use the server certificate otherwise the DFC wont let us write
    gConfigurationData.setOptionInCFG('/DIRAC/Security/UseServerCertificate', 'false')

    gMonitor.addMark("Iteration", 1)

    ############################################################
    #
    # Obtain the files which have not yet been migrated
    #
    self.log.info("Obtaining un-migrated files.")
    res = self.rawIntegrityDB.getUnmigratedFiles()
    if not res['OK']:
      errStr = "Failed to obtain un-migrated files."
      self.log.error(errStr, res['Message'])
      return S_OK()

    # allUnmigratedFilesMeta contain all the files that are not yet
    # migrated (not copied, not registered, not removed), as well as their metadata
    allUnmigratedFilesMeta = res['Value']
    self.log.info("Obtained %s un-migrated files." % len(allUnmigratedFilesMeta))
    if not allUnmigratedFilesMeta:
      return S_OK()

    # activeFiles contains the files that are not yet copied
    activeFiles = {}
    # copiedFiles contains files that are copied but not yet registered
    copiedFiles = {}
    # registeredFiles contains files that are copied, registered, but not removed from source
    registeredFiles = {}

    # Assign them
    for lfn, lfnMetadata in allUnmigratedFilesMeta.items():
      status = lfnMetadata.pop('Status')
      if status == 'Active':
        activeFiles[lfn] = lfnMetadata
      elif status == 'Copied':
        copiedFiles[lfn] = lfnMetadata
      elif status == 'Registered':
        registeredFiles[lfn] = lfnMetadata

    gMonitor.addMark("WaitingFiles", len(activeFiles))
    totalSize = 0
    for lfn, fileDict in activeFiles.items():
      totalSize += int(fileDict['Size'])
      # gMonitor.addMark("TimeInQueue", (fileDict['WaitTime'] / 60))
    gMonitor.addMark("WaitSize", (totalSize / (1024 * 1024 * 1024.0)))

    ############################################################
    #
    # Checking newly copied files
    #

    # Get the list of lfns properly copied and not copied
    filesNewlyCopied, filesNotYetCopied = self.getNewlyCopiedFiles(activeFiles)

    ####################################################
    #
    # Registering copied files
    #
    ####################################################

    filesNewlyRegistered = self.registerCopiedFiles(filesNewlyCopied, copiedFiles,
                                                    allUnmigratedFilesMeta)

    ####################################################
    #
    # Performing the removal from the online storage
    #
    ####################################################
    filesNewlyRemoved = self.removeRegisteredFiles(filesNewlyRegistered, registeredFiles,
                                                   allUnmigratedFilesMeta)

    # Doing some accounting

    migratedSize = sum(allUnmigratedFilesMeta[lfn]['Size'] for lfn in filesNewlyRemoved)

    # The number of files that we failed at migrating
    # is the number of files at the beginning, minus the one we processed completely
    # minus those that are not yet copied
    failedMigrated = len(allUnmigratedFilesMeta) - len(filesNewlyRemoved) - len(filesNotYetCopied)

    res = self.rawIntegrityDB.setLastMonitorTime()
    migratedSizeGB = migratedSize / (1024 * 1024 * 1024.0)
    gMonitor.addMark("TotMigratedSize", migratedSizeGB)
    gMonitor.addMark("NewlyMigrated", len(filesNewlyRemoved))
    gMonitor.addMark("TotMigrated", len(filesNewlyRemoved))
    gMonitor.addMark("FailedMigrated", failedMigrated)
    gMonitor.addMark("TotFailMigrated", failedMigrated)

    return S_OK()



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Transformation Files state machine (LHCb specific)"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from DIRAC.ResourceStatusSystem.PolicySystem.StateMachine import State
from LHCbDIRAC.ProductionManagementSystem.Utilities.StateMachine import LHCbStateMachine

__RCSID__ = "$Id$"


class TransformationFilesStateMachine(LHCbStateMachine):
  """Implementation of the state machine for the TransformationFiles."""

  def __init__(self, state):
    """c'tor Defines the state machine transactions."""

    super(TransformationFilesStateMachine, self).__init__(state)

    self.states = {'Unused-inherited': State(13, ['Unused', 'Processed-inherited', 'MaxReset-inherited'],
                                             defState='Unused-inherited'),
                   'Assigned-inherited': State(12, ['Unused', 'Processed-inherited', 'MaxReset-inherited'],
                                               defState='Assigned-inherited'),
                   'MaxReset-inherited': State(11, ['Unused'], defState='MaxReset-inherited'),
                   'Processed-inherited': State(10),  # final state
                   'Moved': State(9),  # final state
                   'Removed': State(8),  # final state
                   'MissingInFC': State(7),  # final state
                   'NotProcessed': State(6, ['Unused'], defState='NotProcessed'),
                   'ProbInFC': State(5),  # final state
                   'MaxReset': State(4, ['Removed', 'Moved', 'Problematic'], defState='MaxReset'),
                   'Problematic': State(3, ['Removed', 'Unused'], defState='Problematic'),
                   'Processed': State(2),  # final state
                   'Assigned': State(1, ['Unused', 'Processed', 'MaxReset', 'Problematic'],
                                     defState='Assigned'),
                   'Unused': State(0, ['Assigned', 'MissingInFC', 'ProbInFC', 'Problematic',
                                       'Removed', 'NotProcessed', 'Processed', 'Moved'],
                                   defState='Unused')}



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""LHCbDIRAC.AccountingSystem.private.Plotters.JobStepPlotter.

JobStepPlotter.__bases__:
  DIRAC.AccountingSystem.private.Plotters.BaseReporter.BaseReporter
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from DIRAC import S_OK
from DIRAC.AccountingSystem.private.Plotters.BaseReporter import BaseReporter

from LHCbDIRAC.AccountingSystem.Client.Types.JobStep import JobStep

__RCSID__ = "$Id$"


class JobStepPlotter(BaseReporter):
  """JobStepPlotter as an extension of BaseReporter."""

  _typeName = "JobStep"
  _typeKeyFields = [dF[0] for dF in JobStep().definitionKeyFields]

  # .............................................................................
  # CPU Efficiency

  _reportCPUEfficiencyName = "CPU efficiency"

  def _reportCPUEfficiency(self, reportRequest):

    selectField = self._getSelectStringForGrouping(reportRequest['groupingFields'])
    selectFields = (selectField + ", %s, %s, SUM(%s), SUM(%s)",
                    reportRequest['groupingFields'][1] + ['startTime', 'bucketLength', 'CPUTime', 'ExecTime'])

    retVal = self._getTimedData(reportRequest['startTime'],
                                reportRequest['endTime'],
                                selectFields,
                                reportRequest['condDict'],
                                reportRequest['groupingFields'],
                                {'checkNone': True,
                                 'convertToGranularity': 'sum',
                                 'calculateProportionalGauges': False,
                                 'consolidationFunction': self._efficiencyConsolidation})
    if not retVal['OK']:
      return retVal

    dataDict, granularity = retVal['Value']
    self.stripDataField(dataDict, 0)
    if len(dataDict) > 1:
      # Get the total for the plot
      selectFields = ("'Total', %s, %s, SUM(%s),SUM(%s)",
                      ['startTime', 'bucketLength', 'CPUTime', 'ExecTime']
                      )

      retVal = self._getTimedData(reportRequest['startTime'],
                                  reportRequest['endTime'],
                                  selectFields,
                                  reportRequest['condDict'],
                                  reportRequest['groupingFields'],
                                  {'scheckNone': True,
                                   'convertToGranularity': 'sum',
                                   'calculateProportionalGauges': False,
                                   'consolidationFunction': self._efficiencyConsolidation})
      if not retVal['OK']:
        return retVal
      totalDict = retVal['Value'][0]
      self.stripDataField(totalDict, 0)
      for key in totalDict:
        dataDict[key] = totalDict[key]

    return S_OK({'data': dataDict,
                 'granularity': granularity})

  def _plotCPUEfficiency(self, reportRequest, plotInfo, filename):

    metadata = {'title': 'CPU efficiency by %s' % reportRequest['grouping'],
                'starttime': reportRequest['startTime'],
                'endtime': reportRequest['endTime'],
                'span': plotInfo['granularity']}

    return self._generateQualityPlot(filename, plotInfo['data'], metadata)

  # .............................................................................
  # CPU Usage

  _reportCPUUsageName = "CPU time"

  def _reportCPUUsage(self, reportRequest):

    field = 'CPUTime'
    unit = 'time'

    return self.__reportNormPlot(reportRequest, field, unit)

  def _plotCPUUsage(self, reportRequest, plotInfo, filename):

    title = 'CPU usage'

    return self.__plotNormPlot(reportRequest, plotInfo, filename, title)

  # .............................................................................
  # Pie CPU Time

  _reportPieCPUTimeName = 'Pie plot of CPU time'

  def _reportPieCPUTime(self, reportRequest):

    return self.__reportPiePlot(reportRequest, 'CPUTime')

  def _plotPieCPUTime(self, reportRequest, plotInfo, filename):

    title = 'CPU time'
    label = 'CPUTime'

    return self.__plotPie(reportRequest, plotInfo, filename, title, label)

  # .............................................................................
  # Cumulative CPU Time

  _reportCumulativeCPUTimeName = "Cumulative CPU time"

  def _reportCumulativeCPUTime(self, reportRequest):

    field = 'CPUTime'
    unit = 'time'

    return self.__reportCumulativePlot(reportRequest, field, unit)

  def _plotCumulativeCPUTime(self, reportRequest, plotInfo, filename):

    title = 'Cumulative CPU time'

    return self.__plotCumulative(reportRequest, plotInfo, filename, title)

  # .............................................................................
  # Norm CPU Time

  _reportNormCPUTimeName = "NormCPU time"

  def _reportNormCPUTime(self, reportRequest):

    field = 'NormCPUTime'
    unit = 'time'

    return self.__reportNormPlot(reportRequest, field, unit)

  def _plotNormCPUTime(self, reportRequest, plotInfo, filename):

    title = 'Normalized CPU'

    return self.__plotNormPlot(reportRequest, plotInfo, filename, title)

  # .............................................................................
  # Cumulative Norm CPU Time

  _reportCumulativeNormCPUTimeName = "Cumulative normalized CPU time"

  def _reportCumulativeNormCPUTime(self, reportRequest):

    field = 'NormCPUTime'
    unit = 'time'

    return self.__reportCumulativePlot(reportRequest, field, unit)

  def _plotCumulativeNormCPUTime(self, reportRequest, plotInfo, filename):

    title = 'Cumulative Normalized CPU time'

    return self.__plotCumulative(reportRequest, plotInfo, filename, title)

  # .............................................................................
  # Pie Norm CPU Time

  _reportPieNormCPUTimeName = 'Pie plot of NormCPU time'

  def _reportPieNormCPUTime(self, reportRequest):

    return self.__reportPiePlot(reportRequest, 'NormCPUTime')

  def _plotPieNormCPUTime(self, reportRequest, plotInfo, filename):

    title = 'Average Normalized CPU time'
    label = 'NormCPUTime'

    return self.__plotPie(reportRequest, plotInfo, filename, title, label)

  # .............................................................................
  # Input Data

  _reportInputDataName = "Input Data"

  def _reportInputData(self, reportRequest):

    field = 'InputData'
    unit = 'files'

    return self.__reportNormPlot(reportRequest, field, unit)

  def _plotInputData(self, reportRequest, plotInfo, filename):

    title = 'Input data'

    return self.__plotNormPlot(reportRequest, plotInfo, filename, title)

  # .............................................................................
  # Cumulative Input Data

  _reportCumulativeInputDataName = "Cumulative Input Data"

  def _reportCumulativeInputData(self, reportRequest):

    field = 'InputData'
    unit = 'files'

    return self.__reportCumulativePlot(reportRequest, field, unit)

  def _plotCumulativeInputData(self, reportRequest, plotInfo, filename):

    title = 'Cumulative Input Data'

    return self.__plotCumulative(reportRequest, plotInfo, filename, title)

  # .............................................................................
  # Pie Input Data

  _reportPieInputDataName = 'Pie plot of Input Data'

  def _reportPieInputData(self, reportRequest):

    return self.__reportPiePlot(reportRequest, 'InputData')

  def _plotPieInputData(self, reportRequest, plotInfo, filename):

    title = 'Pie plot of Input data'
    label = 'InputData'

    return self.__plotPie(reportRequest, plotInfo, filename, title, label)

  # .............................................................................
  # Output Data

  _reportOutputDataName = "Output Data"

  def _reportOutputData(self, reportRequest):

    field = 'OutputData'
    units = 'files'

    return self.__reportNormPlot(reportRequest, field, units)

  def _plotOutputData(self, reportRequest, plotInfo, filename):

    title = 'Output data'

    return self.__plotNormPlot(reportRequest, plotInfo, filename, title)

  # .............................................................................
  # Cumulative Output Data

  _reportCumulativeOutputDataName = "Cumulative OutputData"

  def _reportCumulativeOutputData(self, reportRequest):

    field = 'OutputData'
    unit = 'files'

    return self.__reportCumulativePlot(reportRequest, field, unit)

  def _plotCumulativeOutputData(self, reportRequest, plotInfo, filename):

    title = 'Cumulative Output Data'

    return self.__plotCumulative(reportRequest, plotInfo, filename, title)

  # .............................................................................
  # Pie Output Data

  _reportPieOutputDataName = 'Pie plot of Output Data'

  def _reportPieOutputData(self, reportRequest):

    return self.__reportPiePlot(reportRequest, 'OutputData')

  def _plotPieOutputData(self, reportRequest, plotInfo, filename):

    title = 'Pie plot of  Output data'
    label = 'OutputData'

    return self.__plotPie(reportRequest, plotInfo, filename, title, label)

  # .............................................................................
  # Input Events

  _reportInputEventsName = "Input Events"

  def _reportInputEvents(self, reportRequest):
    return self.__reportNumberOfField(reportRequest, 'InputEvents')

  def _plotInputEvents(self, reportRequest, plotInfo, filename):

    title = 'InputEvents'
    label = 'Input Events'

    return self.__plotNumberOfField(reportRequest, plotInfo, filename, title, label)

  # .............................................................................
  # Cumulative Input Events

  _reportCumulativeInputEventsName = "Cumulative Input Events"

  def _reportCumulativeInputEvents(self, reportRequest):
    return self.__reportCumulativeNumberOfField(reportRequest, 'InputEvents')

  def _plotCumulativeInputEvents(self, reportRequest, plotInfo, filename):

    title = 'Cumulative Input Events '
    label = 'Input Events'

    return self.__plotCumulativeNumberOfField(reportRequest, plotInfo, filename, title, label)

  # .............................................................................
  # Pie Input Events

  _reportPieInputEventsName = 'Pie plot of Input Events'

  def _reportPieInputEvents(self, reportRequest):
    return self.__reportPiePlot(reportRequest, 'InputEvents')

  def _plotPieInputEvents(self, reportRequest, plotInfo, filename):

    title = 'Pie plot of Input Events'
    label = 'InputEvents'

    return self.__plotPie(reportRequest, plotInfo, filename, title, label)

  # .............................................................................
  # Output Events

  _reportOutputEventsName = "Output Events"

  def _reportOutputEvents(self, reportRequest):
    return self.__reportNumberOfField(reportRequest, 'OutputEvents')

  def _plotOutputEvents(self, reportRequest, plotInfo, filename):

    title = 'OutputEvents'
    label = 'Output Events'

    return self.__plotNumberOfField(reportRequest, plotInfo, filename, title, label)

  # .............................................................................
  # Cumulative Output Events

  _reportCumulativeOutputEventsName = "Cumulative Output Events"

  def _reportCumulativeOutputEvents(self, reportRequest):
    return self.__reportCumulativeNumberOfField(reportRequest, 'OutputEvents')

  def _plotCumulativeOutputEvents(self, reportRequest, plotInfo, filename):

    title = 'Cumulative Output Events '
    label = 'Output Events'

    return self.__plotCumulativeNumberOfField(reportRequest, plotInfo, filename, title, label)

  # .............................................................................
  # Pie Output Events

  _reportPieOutputEventsName = 'Pie plot of Output Events'

  def _reportPieOutputEvents(self, reportRequest):
    return self.__reportPiePlot(reportRequest, 'OutputEvents')

  def _plotPieOutputEvents(self, reportRequest, plotInfo, filename):

    title = 'Pie plot of Output Events'
    label = 'OutputEvents'

    return self.__plotPie(reportRequest, plotInfo, filename, title, label)

  # .............................................................................
  # Input Events Per Output Events

  _reportInputEventsPerOutputEventsName = 'Input/Output Events'

  def _reportInputEventsPerOutputEvents(self, reportRequest):
    return self.__report2D(reportRequest, 'InputEvents', 'OutputEvents')

  def _plotInputEventsPerOutputEvents(self, reportRequest, plotInfo, filename):
    return self.__plot2D(reportRequest, plotInfo, filename, "Input/Output Events")

  # .............................................................................
  # CPU Time Per Output Events

  _reportCPUTimePerOutputEventsName = 'CPUTime/Output Events'

  def _reportCPUTimePerOutputEvents(self, reportRequest):
    return self.__report2D(reportRequest, 'CPUTime', 'OutputEvents')

  def _plotCPUTimePerOutputEvents(self, reportRequest, plotInfo, filename):
    return self.__plot2D(reportRequest, plotInfo, filename, "CPUTime/Output Events")

  # .............................................................................
  # CPU Time Per Input Events

  _reportCPUTimePerInputEventsName = 'CPUTime/Input Events'

  def _reportCPUTimePerInputEvents(self, reportRequest):
    return self.__report2D(reportRequest, 'CPUTime', 'InputEvents')

  def _plotCPUTimePerInputEvents(self, reportRequest, plotInfo, filename):
    return self.__plot2D(reportRequest, plotInfo, filename, "CPUTime/Input Events")

  # .............................................................................
  # HELPER methods

  def __reportNormPlot(self, reportRequest, field, unit):

    selectField = self._getSelectStringForGrouping(reportRequest['groupingFields'])
    selectFields = (selectField + ", %s, %s, SUM(%s)",
                    reportRequest['groupingFields'][1] + ['startTime', 'bucketLength', field]
                    )

    startTime = reportRequest['startTime']
    endTime = reportRequest['endTime']

    retVal = self._getTimedData(startTime,
                                endTime,
                                selectFields,
                                reportRequest['condDict'],
                                reportRequest['groupingFields'],
                                {})
    if not retVal['OK']:
      return retVal

    dataDict, granularity = retVal['Value']
    self.stripDataField(dataDict, 0)

    # 2nd element ( maxValue ) not used
    dataDict, __ = self._divideByFactor(dataDict, granularity)
    dataDict = self._fillWithZero(granularity, startTime, endTime, dataDict)

    accumMaxVal = self._getAccumulationMaxValue(dataDict)
    suitableUnits = self._findSuitableRateUnit(dataDict, accumMaxVal, unit)

    # 3rd element ( maxValue ) not used
    baseDataDict, graphDataDict, __, unitName = suitableUnits

    return S_OK({'data': baseDataDict,
                 'graphDataDict': graphDataDict,
                 'granularity': granularity,
                 'unit': unitName})

  def __plotNormPlot(self, reportRequest, plotInfo, filename, title):

    metadata = {'title': '%s by %s' % (title, reportRequest['grouping']),
                'starttime': reportRequest['startTime'],
                'endtime': reportRequest['endTime'],
                'span': plotInfo['granularity'],
                'ylabel': plotInfo['unit']}

    return self._generateStackedLinePlot(filename, plotInfo['graphDataDict'], metadata)

  def __reportPiePlot(self, reportRequest, field):
    # selectFields = (self._getSelectStringForGrouping(reportRequest[ 'groupingFields' ]) + ", SUM(%s/%s)",
    #                 reportRequest[ 'groupingFields' ][1] + [ field, 'entriesInBucket'
    #                               ]
    #               )

    selectField = self._getSelectStringForGrouping(reportRequest['groupingFields'])
    selectFields = (selectField + ", SUM(%s)",
                    reportRequest['groupingFields'][1] + [field])

    retVal = self._getSummaryData(reportRequest['startTime'],
                                  reportRequest['endTime'],
                                  selectFields,
                                  reportRequest['condDict'],
                                  reportRequest['groupingFields'],
                                  {})
    if not retVal['OK']:
      return retVal
    dataDict = retVal['Value']
    # bins = self._getBins(self._typeName, reportRequest[ 'startTime' ], reportRequest[ 'endTime' ])
    # numBins = len(bins)
    # for key in dataDict:
    #   dataDict[ key ] = float(dataDict[ key ] / numBins)
    return S_OK({'data': dataDict})

  def __plotPie(self, reportRequest, plotInfo, filename, title, label):
    metadata = {'title': '%s by %s' % (title, reportRequest['grouping']),
                'ylabel': label,
                'starttime': reportRequest['startTime'],
                'endtime': reportRequest['endTime']}
    return self._generatePiePlot(filename, plotInfo['data'], metadata)

  def __reportCumulativePlot(self, reportRequest, field, unit):

    selectField = self._getSelectStringForGrouping(reportRequest['groupingFields'])
    selectFields = (selectField + ", %s, %s, SUM(%s)",
                    reportRequest['groupingFields'][1] + ['startTime', 'bucketLength', field]
                    )

    startTime = reportRequest['startTime']
    endTime = reportRequest['endTime']

    retVal = self._getTimedData(startTime,
                                endTime,
                                selectFields,
                                reportRequest['condDict'],
                                reportRequest['groupingFields'],
                                {})
    if not retVal['OK']:
      return retVal

    dataDict, granularity = retVal['Value']
    self.stripDataField(dataDict, 0)

    dataDict = self._fillWithZero(granularity, startTime, endTime, dataDict)
    dataDict = self._accumulate(granularity, startTime, endTime, dataDict)

    accumMaxValue = self._getAccumulationMaxValue(dataDict)
    suitableUnits = self._findSuitableUnit(dataDict, accumMaxValue, unit)

    # 3rd element ( maxValue ) unused
    baseDataDict, graphDataDict, __, unitName = suitableUnits

    return S_OK({'data': baseDataDict,
                 'graphDataDict': graphDataDict,
                 'granularity': granularity,
                 'unit': unitName})

  def __plotCumulative(self, reportRequest, plotInfo, filename, title):

    metadata = {'title': '%s by %s' % (title, reportRequest['grouping']),
                'starttime': reportRequest['startTime'],
                'endtime': reportRequest['endTime'],
                'span': plotInfo['granularity'],
                'ylabel': plotInfo['unit'],
                'sort_labels': 'last_value'}

    return self._generateCumulativePlot(filename, plotInfo['graphDataDict'], metadata)

  def __reportNumberOfField(self, reportRequest, field):

    selectField = self._getSelectStringForGrouping(reportRequest['groupingFields'])
    selectFields = (selectField + ", %s, %s, SUM(%s)",
                    reportRequest['groupingFields'][1] + ['startTime', 'bucketLength', field])

    startTime = reportRequest['startTime']
    endTime = reportRequest['endTime']

    retVal = self._getTimedData(startTime,
                                endTime,
                                selectFields,
                                reportRequest['condDict'],
                                reportRequest['groupingFields'],
                                {})
    if not retVal['OK']:
      return retVal

    dataDict, granularity = retVal['Value']
    self.stripDataField(dataDict, 0)

    # 2nd element ( maxValue ) unused
    dataDict, __ = self._divideByFactor(dataDict, granularity)
    dataDict = self._fillWithZero(granularity, startTime, endTime, dataDict)

    accumMaxValue = self._getAccumulationMaxValue(dataDict)
    suitableUnits = self._findSuitableRateUnit(dataDict, accumMaxValue, "files")

    # 3rd element ( maxValue ) unused
    baseDataDict, graphDataDict, __, unitName = suitableUnits

    return S_OK({'data': baseDataDict,
                 'graphDataDict': graphDataDict,
                 'granularity': granularity,
                 'unit': unitName})

  def __plotNumberOfField(self, reportRequest, plotInfo, filename, title, label):

    startEpoch = reportRequest['startTime']
    endEpoch = reportRequest['endTime']
    granularity = plotInfo['granularity']
    dataDict = plotInfo['data']

    metadata = {'title': '%s  by %s' % (title, reportRequest['grouping']),
                'starttime': startEpoch,
                'endtime': endEpoch,
                'span': granularity,
                'skipEdgeColor': True,
                'ylabel': label}

    dataDict = self._fillWithZero(granularity, startEpoch, endEpoch, dataDict)
    return self._generateStackedLinePlot(filename, dataDict, metadata)

  def __reportCumulativeNumberOfField(self, reportRequest, field):

    selectField = self._getSelectStringForGrouping(reportRequest['groupingFields'])
    selectFields = (selectField + ", %s, %s, SUM(%s)",
                    reportRequest['groupingFields'][1] + ['startTime', 'bucketLength', field]
                    )

    startTime = reportRequest['startTime']
    endTime = reportRequest['endTime']

    retVal = self._getTimedData(startTime,
                                endTime,
                                selectFields,
                                reportRequest['condDict'],
                                reportRequest['groupingFields'],
                                {'convertToGranularity': 'average', 'checkNone': True})
    if not retVal['OK']:
      return retVal

    dataDict, granularity = retVal['Value']
    self.stripDataField(dataDict, 0)

    dataDict = self._fillWithZero(granularity, startTime, endTime, dataDict)
    dataDict = self._accumulate(granularity, startTime, endTime, dataDict)

    return S_OK({'data': dataDict,
                 'granularity': granularity})

  def __plotCumulativeNumberOfField(self, reportRequest, plotInfo, filename, title, label):

    startEpoch = reportRequest['startTime']
    endEpoch = reportRequest['endTime']
    granularity = plotInfo['granularity']
    dataDict = plotInfo['data']

    metadata = {'title': '%s  by %s' % (title, reportRequest['grouping']),
                'starttime': startEpoch,
                'endtime': endEpoch,
                'span': granularity,
                'skipEdgeColor': True,
                'ylabel': label}

    dataDict = self._fillWithZero(granularity, startEpoch, endEpoch, dataDict)
    return self._generateStackedLinePlot(filename, dataDict, metadata)

  # FIXME: smells like unused
  def __reportAverageNumberOfField(self, reportRequest, field):
    # selectFields = (self._getSelectStringForGrouping(reportRequest[ 'groupingFields' ]) + ",SUM(%s/%s)",
    #                 reportRequest[ 'groupingFields' ][1] + [Field, 'entriesInBucket'
    #                               ]
    #               )

    selectField = self._getSelectStringForGrouping(reportRequest['groupingFields'])
    selectFields = (selectField + ",SUM(%s)",
                    reportRequest['groupingFields'][1] + [field])

    retVal = self._getSummaryData(reportRequest['startTime'],
                                  reportRequest['endTime'],
                                  selectFields,
                                  reportRequest['condDict'],
                                  reportRequest['groupingFields'],
                                  {})
    if not retVal['OK']:
      return retVal

    dataDict = retVal['Value']
    # bins = self._getBins(self._typeName, reportRequest[ 'startTime' ], reportRequest[ 'endTime' ])
    # numBins = len(bins)
    # for key in dataDict:
    #  dataDict[ key ] = float(dataDict[ key ] / numBins)
    return S_OK({'data': dataDict})

  # FIXME: smells like unused
  def __plotAverageNumberOfField(self, reportRequest, plotInfo, filename, title, label):

    metadata = {'title': '%s by %s' % (title, reportRequest['grouping']),
                'ylabel': label,
                'starttime': reportRequest['startTime'],
                'endtime': reportRequest['endTime']}
    return self._generatePiePlot(filename, plotInfo['data'], metadata)

  def __report2D(self, reportRequest, field1, field2):

    selectField = self._getSelectStringForGrouping(reportRequest['groupingFields'])
    selectFields = (selectField + ", %s, %s, SUM(%s), SUM(%s)",
                    reportRequest['groupingFields'][1] + ['startTime', 'bucketLength',
                                                          field1, field2])

    startTime = reportRequest['startTime']
    endTime = reportRequest['endTime']

    retVal = self._getTimedData(startTime,
                                endTime,
                                selectFields,
                                reportRequest['condDict'],
                                reportRequest['groupingFields'],
                                {'checkNone': True,
                                 'convertToGranularity': 'sum',
                                 'calculateProportionalGauges': True})
    if not retVal['OK']:
      return retVal

    dataDict, granularity = retVal['Value']
    self.stripDataField(dataDict, 0)

    dataDict = self._fillWithZero(granularity, startTime, endTime, dataDict)
    return S_OK({'data': dataDict,
                 'granularity': granularity})

  def __plot2D(self, reportRequest, plotInfo, filename, label):

    metadata = {'title': 'Jobs by %s' % reportRequest['grouping'],
                'starttime': reportRequest['startTime'],
                'endtime': reportRequest['endTime'],
                'span': plotInfo['granularity'],
                'ylabel': label}

    return self._generateTimedStackedBarPlot(filename, plotInfo['data'], metadata)



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Module to upload specified job output files according to the parameters
defined in the user workflow."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import os
import random
import re

from DIRAC import S_OK, S_ERROR, gLogger, gConfig
from DIRAC.Core.Utilities.File import getGlobbedFiles
from DIRAC.DataManagementSystem.Client.FailoverTransfer import FailoverTransfer
from DIRAC.DataManagementSystem.Utilities.DMSHelpers import resolveSEGroup

from LHCbDIRAC.Core.Utilities.ProductionData import constructUserLFNs
from LHCbDIRAC.Workflow.Modules.ModuleBase import ModuleBase
from LHCbDIRAC.Core.Utilities.ResolveSE import getDestinationSEList

__RCSID__ = "$Id$"


class UserJobFinalization(ModuleBase):
  """Finalization of user jobs."""

  #############################################################################
  def __init__(self, bkClient=None, dm=None):
    """Module initialization."""

    self.log = gLogger.getSubLogger("UserJobFinalization")
    super(UserJobFinalization, self).__init__(self.log, bkClientIn=bkClient, dm=dm)

    self.version = __RCSID__
    self.enable = True
    self.defaultOutputSE = resolveSEGroup(gConfig.getValue('/Resources/StorageElementGroups/Tier1-USER', []))
    self.failoverSEs = resolveSEGroup(gConfig.getValue('/Resources/StorageElementGroups/Tier1-Failover', []))
    # List all parameters here
    self.request = None
    # Always allow any files specified by users
    self.outputDataFileMask = ''
    self.userOutputData = []
    self.userOutputSE = ''
    self.userOutputPath = ''
    self.failoverTransfer = None
    self.replicateUserOutputData = False
    self.userPrependString = ''

  #############################################################################
  def _resolveInputVariables(self):
    """By convention the module parameters are resolved here."""
    super(UserJobFinalization, self)._resolveInputVariables()

    # Use LHCb utility for local running via dirac-jobexec
    if 'UserOutputData' in self.workflow_commons:
      userOutputData = self.workflow_commons['UserOutputData']
      if not isinstance(userOutputData, list):
        userOutputData = [i.strip() for i in userOutputData.split(';')]
      self.userOutputData = userOutputData

    if 'UserOutputSE' in self.workflow_commons:
      specifiedSE = self.workflow_commons['UserOutputSE']
      if not isinstance(specifiedSE, list):
        self.userOutputSE = [i.strip() for i in specifiedSE.split(';')]
    else:
      self.log.verbose('No UserOutputSE specified, using default value: %s' % (', '.join(self.defaultOutputSE)))
      self.userOutputSE = []

    if 'UserOutputPath' in self.workflow_commons:
      self.userOutputPath = self.workflow_commons['UserOutputPath']

    if 'ReplicateUserOutputData' in self.workflow_commons and self.workflow_commons['ReplicateUserOutputData']:
      self.replicateUserOutputData = True

    if 'UserOutputLFNPrepend' in self.workflow_commons:
      self.userPrependString = self.workflow_commons['UserOutputLFNPrepend']

  #############################################################################

  def execute(self, production_id=None, prod_job_id=None, wms_job_id=None,
              workflowStatus=None, stepStatus=None,
              wf_commons=None, step_commons=None,
              step_number=None, step_id=None, orderedSEs=None):
    """Main execution function."""

    try:

      super(UserJobFinalization, self).execute(self.version, production_id, prod_job_id, wms_job_id,
                                               workflowStatus, stepStatus,
                                               wf_commons, step_commons, step_number, step_id)

      self._resolveInputVariables()

      # Earlier modules may have populated the report objects
      self.request.RequestName = 'job_%d_request.xml' % self.jobID
      self.request.JobID = self.jobID
      self.request.SourceComponent = "Job_%d" % self.jobID

      if not self._checkWFAndStepStatus():
        return S_OK()

      if not self.userOutputData:
        self.log.info("No user output data is specified for this job, nothing to do")
        return S_OK("No output data to upload")

      self.log.info("User specified output file list is: %s" % (', '.join(self.userOutputData)))

      globList = []
      for i in self.userOutputData:
        if re.search(r'\*', i):
          globList.append(i)

      # Check whether list of userOutputData is a globbable pattern
      if globList:
        for i in globList:
          self.userOutputData.remove(i)

        globbedOutputList = list(set(getGlobbedFiles(globList)))
        if globbedOutputList:
          self.log.info('Found a pattern in the output data file list, \
          extra files to upload are: %s' % (', '.join(globbedOutputList)))
          self.userOutputData += globbedOutputList
        else:
          self.log.info("No files were found on the local disk for the following patterns: %s" % (', '.join(globList)))

      self.log.info("Final list of files to upload are: %s" % (', '.join(self.userOutputData)))

      # Determine the final list of possible output files for the workflow
      # and all the parameters needed to upload them.
      outputList = []
      for i in self.userOutputData:
        outputList.append({'outputDataType': ('.'.split(i)[-1]).upper(),
                           'outputDataName': os.path.basename(i)})

      userOutputLFNs = []
      if self.userOutputData:
        self.log.info("Constructing user output LFN(s) for %s" % (', '.join(self.userOutputData)))

        userOutputLFNs = constructUserLFNs(self.jobID, self._getCurrentOwner(),
                                           self.userOutputData, self.userOutputPath,
                                           self.userPrependString)

      self.log.verbose("Calling getCandidateFiles( %s, %s, %s)" % (outputList, userOutputLFNs,
                                                                   self.outputDataFileMask))
      try:
        fileDict = self.getCandidateFiles(outputList, userOutputLFNs, self.outputDataFileMask)
      except os.error as e:
        self.setApplicationStatus(e)
        return S_OK()

      try:
        fileMetadata = self.getFileMetadata(fileDict)
      except RuntimeError as e:
        self.setApplicationStatus(e)
        return S_OK()

      if not fileMetadata:
        self.log.info("No output data files were determined to be uploaded for this workflow")
        self.setApplicationStatus('No Output Data Files To Upload')
        return S_OK()

      if not orderedSEs:
        orderedSEs = self._getOrderedSEsList()

      self.log.info("Ordered list of output SEs is: %s" % (', '.join(orderedSEs)))
      final = {}
      for fileName, metadata in fileMetadata.items():
        final[fileName] = metadata
        final[fileName]['resolvedSE'] = orderedSEs

      # At this point can exit and see exactly what the module will upload
      if not self._enableModule():
        self.log.info("Module disabled would have attempted to upload the files %s" % ', '.join(final))
        for fileName, metadata in final.items():
          self.log.info('--------%s--------' % fileName)
          for n, v in metadata.items():
            self.log.info('%s = %s' % (n, v))

        return S_OK("Module is disabled by control flag")

      # Disable the watchdog check in case the file uploading takes a long time
      self._disableWatchdogCPUCheck()

      # Instantiate the failover transfer client with the global request object
      if not self.failoverTransfer:
        self.failoverTransfer = FailoverTransfer(self.request)

      # One by one upload the files with failover if necessary
      replication = {}
      failover = {}
      uploaded = []
      for fileName, metadata in final.items():
        self.log.info("Attempting to store %s to the following SE(s): %s" % (fileName,
                                                                             ', '.join(metadata['resolvedSE'])))
        fileMetaDict = {'Size': metadata['filedict']['Size'],
                        'LFN': metadata['filedict']['LFN'],
                        'GUID': metadata['filedict']['GUID'],
                        'Checksum': metadata['filedict']['Checksum'],
                        'ChecksumType': metadata['filedict']['ChecksumType']}
        result = self.failoverTransfer.transferAndRegisterFile(fileName=fileName,
                                                               localPath=metadata['localpath'],
                                                               lfn=metadata['filedict']['LFN'],
                                                               destinationSEList=metadata['resolvedSE'],
                                                               fileMetaDict=fileMetaDict,
                                                               masterCatalogOnly=True)
        if not result['OK']:
          self.log.error("Could not transfer and register %s with metadata:\n %s" % (fileName, metadata))
          failover[fileName] = metadata
        else:
          # Only attempt replication after successful upload
          lfn = metadata['lfn']
          uploaded.append(lfn)
          seList = metadata['resolvedSE']
          replicateSE = ''
          uploadedSE = result['Value'].get('uploadedSE', '')
          if uploadedSE:
            for se in seList:
              if not se == uploadedSE:
                replicateSE = se
                break

          if replicateSE and lfn and self.replicateUserOutputData:
            self.log.info("Will attempt to replicate %s to %s" % (lfn, replicateSE))
            replication[lfn] = (uploadedSE, replicateSE, fileMetaDict)

      cleanUp = False
      for fileName, metadata in failover.items():
        random.shuffle(self.failoverSEs)
        targetSE = metadata['resolvedSE'][0]
        if len(metadata['resolvedSE']) > 1:
          replicateSE = metadata['resolvedSE'][1]
        else:
          replicateSE = ''
        metadata['resolvedSE'] = self.failoverSEs
        fileMetaDict = {'Size': metadata['filedict']['Size'],
                        'LFN': metadata['filedict']['LFN'],
                        'GUID': metadata['filedict']['GUID']}
        result = self.failoverTransfer.transferAndRegisterFileFailover(fileName,
                                                                       metadata['localpath'],
                                                                       metadata['lfn'],
                                                                       targetSE,
                                                                       metadata['resolvedSE'],
                                                                       fileMetaDict=fileMetaDict,
                                                                       masterCatalogOnly=True)
        if not result['OK']:
          self.log.error("Could not transfer and register %s with metadata:\n %s" % (fileName, metadata))
          cleanUp = True
          continue  # for users can continue even if one completely fails
        else:
          lfn = metadata['lfn']
          uploaded.append(lfn)
          # Even when using Failover, one needs to replicate to a second SE
          if replicateSE and self.replicateUserOutputData:
            replication[lfn] = (targetSE, replicateSE, fileMetaDict)

      # For files correctly uploaded must report LFNs to job parameters
      if uploaded:
        report = ', '.join(uploaded)
        self.setJobParameter('UploadedOutputData', report)

      # Now after all operations, retrieve potentially modified request object
      self.request = self.failoverTransfer.request

      # If some or all of the files failed to be saved to failover
      if cleanUp:
        self.workflow_commons['Request'] = self.request
        # Leave any uploaded files just in case it is useful for the user
        # do not try to replicate any files.
        return S_ERROR("Failed To Upload Output Data")

      for lfn, (uploadedSE, repSE, fileMetaDictItem) in replication.items():
        self.failoverTransfer._setFileReplicationRequest(lfn, repSE, fileMetaDictItem, uploadedSE)

      self.workflow_commons['Request'] = self.failoverTransfer.request

      self.generateFailoverFile()

      self.setApplicationStatus("Job Finished Successfully")

      return S_OK('Output data uploaded')

    except Exception as e:  # pylint:disable=broad-except
      self.log.exception("Failure in UserJobFinalization execute module", lException=e)
      self.setApplicationStatus(repr(e))
      return S_ERROR(str(e))

    finally:
      super(UserJobFinalization, self).finalize(self.version)

  #############################################################################

  def _getOrderedSEsList(self):
    """Returns list of ordered SEs to which trying to upload."""
    # FIXME: remove all banned SEs (not the force ones)
    # First get the local (or assigned) SE to try first for upload and others in random fashion
    localSEs = set(getDestinationSEList('Tier1-USER', self.siteName, outputmode='local'))
    self.log.verbose("Site Local SE for user outputs is: %s" % (list(localSEs)))
    userSEs = set(self.userOutputSE)
    otherSEs = set(self.defaultOutputSE) - localSEs - userSEs
    # If a user SE is  local set it first
    topSEs = userSEs & localSEs
    # reordered user SEs, setting local first
    userSEs = list(topSEs) + list(userSEs - topSEs)
    localSEs = list(localSEs - topSEs)
    if len(userSEs) < 2 and localSEs:
      # Set a local SE first
      orderedSEs = localSEs[0:1] + userSEs + localSEs[1:]
    else:
      orderedSEs = userSEs + localSEs
    random.shuffle(list(otherSEs))
    orderedSEs += otherSEs

    return orderedSEs
